package datavisualizer;

import javax.swing.JInternalFrame;
import java.util.Vector;
import javawrapper.ModelException;

/**
* This class manages a single data file for the data visualizer.  It is
* responsible for parsing the data out of the file, telling the data
* visualizer what charts are available, and providing requested charts.
*
* Each DataFileManager object keeps track of the charts it has created and
* placed in JInternalFrame objects.  These objects are under the direct
* management of the entitity to which they were originally passed when
* DrawChart() was called to create them, but the DataFileManager object
* retains the ability to update them per user request.
* <p>Copyright: Copyright (c) Charles D. Canham 2004</p>
* <p>Company: Institute of Ecosystem Studies</p>
* @author Lora E. Murphy
* @version 1.0
*
* <br>Edit history:
* <br>------------------
* <br>April 28, 2004: Submitted in beta version (LEM)
* <br>November 18, 2004:  Added real-time data visualization (LEM)
*/

abstract public class DataFileManager {
  protected String m_sFilename; /**<File under management*/
  protected Vector mp_oCharts; /**<All charts under management*/
  protected Legend m_oLegend; /**<The legend for this file*/

  /**
   * Get the legend for this file.
   * @return The legend.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public Legend GetLegend() {
    return m_oLegend;
  }

  /**
   * Gets the number of open charts for this file.
   * @return int Number of open charts for this file.
   */
  public int GetNumberOpenCharts() {
    return mp_oCharts.size();
  }

  /**
   * Closes all the open charts for this window.
   */
  public void CloseAllCharts() {
    JInternalFrame jFrame = null;
    int i, iNumOpenCharts = mp_oCharts.size();

    if (iNumOpenCharts == 0) return;

    for (i = iNumOpenCharts-1; i >= 0; i--) {
      Object oElement = mp_oCharts.elementAt(i);
      if (oElement != null) {
        jFrame = (JInternalFrame) oElement;
        jFrame.setVisible(false);
        jFrame.dispose();
        mp_oCharts.removeElementAt(i);
      }
    }

    m_oLegend.setVisible(false);
    m_oLegend.dispose();
  }

  /**
   * Any clean-up tasks that must be done can be put here.  This will be
   * called before the object is destroyed.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void CleanUp() {

  }

  /**
   * Constructor.
   * @param sFileName  File name for this manager to manage.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public DataFileManager(String sFileName) {
    m_sFilename = sFileName;
    mp_oCharts = new Vector(0);
  }

  /**
   * Gets the file being managed.
   * @return File name of managed file.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public String GetFileName() {
    return m_sFilename;
  }

  /**
   * Gets a list of the line graphs this object is capable of drawing
   * @return List of line graph labels
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  abstract public String[] GetLineGraphOptions();

  /**
   * Gets a list of maps this object is capable of drawing
   * @return List of map labels
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  abstract public String[] GetMapOptions();

  /**
   * Gets a list of histograms this object is capable of drawing
   * @return List of histogram labels
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  abstract public String[] GetHistogramOptions();

  /**
   * Gets a list of tables this object is capable of drawing
   * @return List of table labels
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  abstract public String[] GetTableOptions();

  /**
       * Creates a requested chart as a JInternalFrame.  The JInternalFrame will not
   * be packed or set to visible - it will simply have its GUI components.  If
   * the chart requested already exists, the existing chart is returned.
   * @param sGraphName The string name of the chart - should be from one of the
   * "GetXOptions()" methods
   * @return The chart window, or null if no such chart can be drawn.
   * @throws ModelException if there are any problems drawing the chart.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  abstract protected JInternalFrame DrawChart(String sGraphName) throws
      ModelException;

  /**
       * Creates a requested chart as a JInternalFrame.  If the chart already exists,
   * a new one is not created; the existing one is returned.  If the chart is
   * new, the JInternalFrame will not be packed or set to visible - it will
   * simply have its GUI components.
   * @param sGraphName The string name of the chart - should be from one of the
   * "GetXOptions()" methods
   * @return The chart window, or null if no such chart can be drawn.
   * @throws ModelException if there are any problems drawing the chart.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public JInternalFrame CreateNewChart(String sGraphName) throws ModelException {

    //Go through existing charts to see if any are already this graph name
    String sTitle;
    int i;
    JInternalFrame oChartWindow;

    for (i = 0; i < mp_oCharts.size(); i++) {

      if (mp_oCharts.elementAt(i) != null) {
        oChartWindow = (JInternalFrame) mp_oCharts.elementAt(i);
        //Get the title and trim off the filename
        sTitle = oChartWindow.getTitle();
        int iPos = sTitle.indexOf(" - ");
        if (iPos > -1) {
          sTitle = sTitle.substring(0, sTitle.indexOf(" - "));
        }
        if (sTitle.equalsIgnoreCase(sGraphName)) {
          return oChartWindow;
        }
      }
    }

    //The old chart did not exist - create a new one
    oChartWindow = DrawChart(sGraphName);

    //Add it to the list of charts under management by this file
    mp_oCharts.add(oChartWindow);

    return oChartWindow;
  }

  /**
   * Prompts the redrawing of all open charts.
   * @throws ModelException wrapping other exceptions
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  abstract public void UpdateCharts() throws ModelException;

  /**
  * Prompts the redrawing of open charts for the current run.  The data
  * file manager should search for and display the most recent information.
  * @throws ModelException if there is a problem drawing the charts.
  */
  abstract public void UpdateCurrentRunCharts() throws ModelException;

}
