package datavisualizer;

import java.awt.event.ActionEvent;
import javax.swing.*;

import java.util.Vector;

import javawrapper.ModelException;

/**
 * This controls the drawing of a histogram for a single piece of grid
 * information.
 * The label of the data is passed in the constructor.  The histogram is
 * defaulted to 10 divisions equally divided between their minimum and maximum
 * values.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>March 29, 2005:  Created (LEM)
 */

public class GridHistogramDataRequest
    extends DataRequest
    implements java.awt.event.ActionListener {

  /**Data member's label, as it appears in the detailed output file*/
  private String m_sLabel;

  /**Name of grid*/
  private String m_sGridName;

  /**Whether our data member being graphed is a float*/
  private boolean m_bIsFloat;

  /**Whether the data being graphed is an integer*/
  private boolean m_bIsInt;

  /**The dataset that will be graphed*/
  protected ModelHistogramDataset m_oDataset;

  /**Field displaying the number of bins in the histogram*/
  private JTextField m_jNumBins = new JTextField("");

  /**Field displaying the bin size in the histogram*/
  private JTextField m_jBinSize = new JTextField("");

  /**Label holding mean value of dataset*/
  private JLabel m_jMean = new JLabel("");

  /**The data member code for this data member*/
  private int m_iDataMemberCode = -1;

  /**The data*/
  private Vector mp_oData = new Vector(0);

  /**Size of the histogram bins*/
  private float m_fBinSize;

  /**Number of bins in the histogram*/
  private int m_iNumBins;

  /**Whether or not to use a logarithmic axis on the Y*/
  private boolean m_bUseLogarithmicAxis;

  /**Whether or not we have to recalculate the bins on chart update (i.e. we
     are charting an unknown data type without defaults)*/
  private boolean m_bRecalcBinsOnUpdate;

  /** Formatter for big numbers */
  private java.text.NumberFormat m_jFormat = java.text.NumberFormat.getInstance();

  /** Formatter for very small numbers */
  java.text.DecimalFormat m_jSmallNumsFormat = new java.text.DecimalFormat(
        "0.#E0");

  /**
   * Constructor.
   * @param sLabel Label of the piece of data, as it would be passed as a data
   * member code.
   * @param bIsFloat Whether this is float data.  If this is set to false,
   * integer data is assumed.
   * @param oManager The parent detailed output file manager.
   * @param sGraphName Name of graph string.
   * @param sGridName Name of grid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>March 29, 2005:  Created (LEM)
   */
  public GridHistogramDataRequest(String sLabel, boolean bIsFloat,
                                  DetailedOutputFileManager oManager,
                                  String sGraphName, String sGridName) {
    super(sGraphName, oManager);

    //Default use of logarithmic axis to true
    m_bUseLogarithmicAxis = true;

    //Set up flags
    m_sLabel = sLabel;
    m_sGridName = sGridName;
    if (bIsFloat) {
      m_bIsFloat = true;
      m_bIsInt = false;
    }
    else {
      m_bIsFloat = false;
      m_bIsInt = true;
    }

    m_oManager = oManager;

    m_iNumBins = 10;
    m_fBinSize = 0;
    m_bRecalcBinsOnUpdate = true;

    //Set up the formatters
    m_jFormat.setMaximumFractionDigits(2);
    m_jFormat.setGroupingUsed(false);
    m_jSmallNumsFormat.setGroupingUsed(false);
  }

  /**
   * Writes the histogram's data to tab-delimited text.
   * @param jOut java.io.FileWriter The file to write to.
   * @throws java.io.IOException if there's a problem writing the file.
   */
  protected void WriteChartDataToFile(java.io.FileWriter jOut) throws java.io.
      IOException {

    //It's possible that if another chart was opened, and this timestep's
    //file read again as a consequence, that the dataset needs to be
    //recreated even if this chart is still displaying correctly
    if (m_oDataset.getSeriesCount() == 0) {
      AddSeriesToDataset();
    }

    float fMean = GetDatasetMean();
    jOut.write("Mean value: " + String.valueOf(fMean) + "\n");

    //Get the number of series - this will be the number of species, with
    //a possible totals series
    int iNumSeries = m_oDataset.getSeriesCount();
    if (iNumSeries <= 0) {
      jOut.write("This histogram has no data.");
      return;
    }

    int i, j;

    //Write out the header columns
    for (i = 1; i <= m_iNumBins; i++) {
      jOut.write("\t" + String.valueOf( (i - 1) * m_fBinSize) + " - " +
                 String.valueOf(i * m_fBinSize));
    }
    jOut.write("\n");

    //Write out the series
    for (i = 0; i < iNumSeries; i++) {
      jOut.write(m_oDataset.getSeriesName(i));
      int iNumItems = m_oDataset.getItemCount(i);
      for (j = 0; j < iNumItems; j++) {
        jOut.write("\t" + String.valueOf(m_oDataset.getYValue(i, j)));
      }
      jOut.write("\n");
    }
  }

  /**
   * Performs actions for the controls in the Histogram window.
   * @param oEvent Event triggered.
   */
  public void actionPerformed(ActionEvent oEvent) {
    try {
      if (oEvent.getSource() instanceof JCheckBox) {
        JCheckBox jBox = (JCheckBox) oEvent.getSource();
        if (jBox.getName().equals("log_checkbox")) {
          if (jBox.isSelected() != m_bUseLogarithmicAxis) {
            m_bUseLogarithmicAxis = jBox.isSelected();
            UpdateChart(m_oManager.GetLegend());
          }
        }
      }
      else if (oEvent.getActionCommand().equals("UpdateBins")) {

        //Make sure the values in the bins are recognizable, greater-than-zero numbers
        int iNumBins = 0;
        float fBinSize = 0;
        try {
          Integer oNumBins = new Integer(m_jNumBins.getText());
          Float oBinSize = new Float(m_jBinSize.getText());
          iNumBins = oNumBins.intValue();
          fBinSize = oBinSize.floatValue();
        }
        catch (java.lang.NumberFormatException oErr) {
          JOptionPane.showMessageDialog(m_oChartFrame,
                                        "The value for number of bins or bin size is not a recognized number.");
          m_jNumBins.setText(String.valueOf(m_iNumBins));
          if (m_fBinSize > 0.01) {
            m_jBinSize.setText(m_jFormat.format(m_fBinSize));
          }
          else {
            m_jBinSize.setText(m_jSmallNumsFormat.format(m_fBinSize));
          }
          return;
        }

        if (iNumBins <= 0 || fBinSize <= 0) {
          JOptionPane.showMessageDialog(m_oChartFrame,
                                        "The values for number of bins and bin size must be greater than zero.");
          m_jNumBins.setText(String.valueOf(m_iNumBins));
          if (m_fBinSize > 0.01) {
            m_jBinSize.setText(m_jFormat.format(m_fBinSize));
          }
          else {
            m_jBinSize.setText(m_jSmallNumsFormat.format(m_fBinSize));
          }
          return;
        }

        if (iNumBins != m_iNumBins || fBinSize != m_fBinSize) {
          m_iNumBins = iNumBins;
          m_fBinSize = fBinSize;

          //Clear the dataset to force a reconstruction
          m_oDataset = null;
          m_oDataset = new ModelHistogramDataset();

          m_bRecalcBinsOnUpdate = false;
          UpdateChart(m_oManager.GetLegend());
        }
      }
      super.actionPerformed(oEvent);
    }
    catch (javawrapper.ModelException oErr) {
      javawrapper.ErrorGUI oHandler = new javawrapper.ErrorGUI(m_oChartFrame);
      oHandler.WriteErrorMessage(oErr);
    }
  }

  /**
   * Gets the bin size for this data request's histogram.
   * @return The bin size.
   *
   * <br>Edit history:
   * <br>------------------
   * March 29, 2005:  Created (LEM)
   */
  public float GetBinSize() {
    return m_fBinSize;
  }

  /**
   * Gets whether or not to use a logarithmic axis.
   * @return True if a logarithmic axis should be used, or false if a linear
   * axis should be used.
   *
   * <br>Edit history:
   * <br>------------------
   * March 29, 2005:  Created (LEM)
   */
  public boolean GetUseLogarithmicAxis() {
    return m_bUseLogarithmicAxis;
  }

  /**
   * Captures the data member code if we're graphing a float.
   * @param sGridName Name of the grid
   * @param sLabel The label of the data member.
   * @param iCode The data member code.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddGridFloatDataMemberCode(String sGridName, String sLabel,
                                         int iCode) {
    if (m_bIsFloat && sGridName.equals(m_sGridName) && sLabel.equals(m_sLabel)) {

      //Bingo - it's this data member that we're graphing
      m_iDataMemberCode = iCode;
    }
  }

  /**
   * Captures the data member code if we're graphing an integer.
   * @param sGridName Name of the grid
   * @param sLabel The label of the data member.
   * @param iCode The data member code.
   *
   * <br>Edit history:
   * <br>------------------
   * March 29, 2005:  Created (LEM)
   */
  public void AddGridIntDataMemberCode(String sGridName, String sLabel,
                                       int iCode) {
    if (m_bIsInt && sGridName.equals(m_sGridName) && sLabel.equals(m_sLabel)) {

      //Bingo - it's this data member that we're graphing
      m_iDataMemberCode = iCode;
    }
  }

  /**
   * Creates the histogram using data accumulated thus far.
   * @param oLegend Legend for this chart.
   * @param sChartTitle Title for this chart.
   * @return The new chart window.
   * @throws datavisualizer.ModelException Passing through underlying
   * exceptions.
   *
   * <br>Edit history:
   * <br>------------------
   * March 29, 2005:  Created (LEM)
   */
  ModelInternalFrame DrawChart(Legend oLegend, String sChartTitle) throws
      ModelException {

    int i;

    //Create our new dataset
    m_oDataset = new ModelHistogramDataset();

    AddSeriesToDataset();
    float fMean = GetDatasetMean();

    //Make the chart
    m_oChartFrame = DataGrapher.DrawHistogram(m_oDataset, m_sLabel,
                                              "Total Number",
                                              sChartTitle, oLegend, this);

    //Add the extra controls to the top of the histogram window
    JPanel jControls = new JPanel(new java.awt.FlowLayout(java.awt.FlowLayout.
        LEFT));

    //Bins boxes
    JLabel jTemp = new JLabel("Number of bins:");
    jTemp.setFont(new javawrapper.ModelFont());
    jControls.add(jTemp);
    m_jNumBins.setFont(new javawrapper.ModelFont());
    m_jNumBins.setText(String.valueOf(m_iNumBins));
    m_jNumBins.setPreferredSize(new java.awt.Dimension(50,
        (int) m_jNumBins.getPreferredSize().getHeight()));
    jControls.add(m_jNumBins);

    jTemp = new JLabel("Bin size:");
    jTemp.setFont(new javawrapper.ModelFont());
    jControls.add(jTemp);
    m_jBinSize.setFont(new javawrapper.ModelFont());

    if (m_fBinSize > 0.01) {
      m_jBinSize.setText(m_jFormat.format(m_fBinSize));
    }
    else {
      m_jBinSize.setText(m_jSmallNumsFormat.format(m_fBinSize));
    }
    m_jBinSize.setPreferredSize(new java.awt.Dimension(75,
        (int) m_jBinSize.getPreferredSize().getHeight()));
    jControls.add(m_jBinSize);

    //Button to change
    JButton jButton = new JButton("Change");
    jButton.setFont(new javawrapper.ModelFont());
    jButton.setActionCommand("UpdateBins");
    jButton.addActionListener(this);
    jControls.add(jButton);

    //Checkbox for logarithmic axis
    JCheckBox jUseLogarithmic = new JCheckBox("Use Logarithmic Axis",
                                              GetUseLogarithmicAxis());
    jUseLogarithmic.setFont(new javawrapper.ModelFont());
    jUseLogarithmic.setName("log_checkbox");
    jUseLogarithmic.addActionListener(this);
    jControls.add(jUseLogarithmic);

    //Display of mean - set under other controls in its own panel
    jTemp = new JLabel("Mean value: ");
    jTemp.setFont(new javawrapper.ModelFont());
    m_jMean.setFont(new javawrapper.ModelFont());
    if (fMean == 0) {
      m_jMean.setText(String.valueOf(fMean));
    }
    else if (fMean > 0.1) {
      m_jMean.setText(m_jFormat.format(fMean));
    }
    else {
      m_jMean.setText(m_jSmallNumsFormat.format(fMean));
    }
    JPanel jMeanPanel = new JPanel();
    jMeanPanel.add(jTemp);
    jMeanPanel.add(m_jMean);

    //Remove the chart, which is the content panel
    org.jfree.chart.ChartPanel jChart = null;
    for (i = 0; i < m_oChartFrame.getContentPane().getComponentCount(); i++) {
      if (m_oChartFrame.getContentPane().getComponent(i) instanceof org.jfree.
          chart.ChartPanel) {
        jChart = (org.jfree.chart.ChartPanel) m_oChartFrame.getContentPane().
            getComponent(i);
      }
    }

    //Recreate the content pane with the controls and the chart
    JPanel jContentPanel = new JPanel(new java.awt.BorderLayout());
    JPanel jTempPanel = new JPanel(new java.awt.BorderLayout());
    jTempPanel.add(jControls, java.awt.BorderLayout.CENTER);
    jTempPanel.add(jMeanPanel, java.awt.BorderLayout.SOUTH);
    jContentPanel.setLayout(new java.awt.BorderLayout());
    jContentPanel.add(jTempPanel, java.awt.BorderLayout.NORTH);
    jContentPanel.add(jChart, java.awt.BorderLayout.CENTER);

    m_oChartFrame.setContentPane(jContentPanel);
    return m_oChartFrame;

  }

  /**
   * Calculates the mean value of all values currently in mp_oData.
   * @return float Mean of all values.
   */
  protected float GetDatasetMean() {
    float fMean = 0;
    int i, iNumItems = mp_oData.size();

    for (i = 0; i < iNumItems; i++) {
      if (mp_oData.elementAt(i) instanceof Float) {
        fMean += ( (Float) mp_oData.elementAt(i)).floatValue();
      } else if (mp_oData.elementAt(i) instanceof Integer) {
        fMean += ( (Integer) mp_oData.elementAt(i)).floatValue();
      }
    }

    if (iNumItems > 0) {
      fMean /= iNumItems;
    }

    return fMean;
  }

  /**
   * This will add all series to the dataset.
   *
   * <br>Edit history:
   * <br>------------------
   * March 29, 2005:  Created (LEM)
   */
  protected void AddSeriesToDataset() {
    double[] p_oNewSeries;
    float fMax = 0;
    int j;

    //If our binsize is 0, figure it out from the data.  Get the maximum value
    //in the dataset.  We'll want to refigure this at every update because we
    //don't know the final limits of whatever we're graphing.
    if (m_bRecalcBinsOnUpdate) {
      //Get the maximum value
      for (j = 0; j < mp_oData.size(); j++) {
        Number oNumber = (Number) mp_oData.elementAt(j);
        if (oNumber.floatValue() > fMax) {
          fMax = oNumber.floatValue();
        }
      }

      m_fBinSize = (fMax * (float) 1.1) / m_iNumBins;
    }

    //Add all accumulated data to the dataset
    p_oNewSeries = new double[mp_oData.size()];

    for (j = 0; j < p_oNewSeries.length; j++) {
      if (m_bIsFloat) {
        Float fVal = (Float) mp_oData.elementAt(j);
        p_oNewSeries[j] = fVal.doubleValue();
      }
      else {
        Integer iVal = (Integer) mp_oData.elementAt(j);
        p_oNewSeries[j] = iVal.doubleValue();
      }
    }

    if (p_oNewSeries.length > 0) {
      m_oDataset.addSeries(m_sLabel, p_oNewSeries,
                           m_iNumBins, m_fBinSize);
    }
  }

  /**
   * Updates the histogram.
   * @param oLegend Legend for this chart.
   * @throws datavisualizer.ModelException Passing through an underlying
   * exception.
   *
   * <br>Edit history:
   * <br>------------------
   * March 29, 2005:  Created (LEM)
   */
  void UpdateChart(Legend oLegend) throws ModelException {

    //If there are no series in the dataset, add them now (in case the
    //file was re-parsed while this chart was open)
    if (m_oDataset.getSeriesCount() == 0) {
      AddSeriesToDataset();
    }
    float fMean = GetDatasetMean();

    DataGrapher.UpdateHistogram(m_oDataset, m_oChartFrame, oLegend,
                                this);

    //Update bin information, if we're updating
    if (m_bRecalcBinsOnUpdate) {
      m_jNumBins.setText(m_jFormat.format(m_iNumBins));
      if (m_fBinSize > 0.01) {
        m_jBinSize.setText(m_jFormat.format(m_fBinSize));
      }
      else {
        m_jBinSize.setText(m_jSmallNumsFormat.format(m_fBinSize));
      }
    }

    //Update mean value
    if (fMean == 0) {
      m_jMean.setText(String.valueOf(fMean));
    }
    else if (fMean > 0.1) {
      m_jMean.setText(m_jFormat.format(fMean));
    }
    else {
      m_jMean.setText(m_jSmallNumsFormat.format(fMean));
    }
  }

  /**
   * If our data piece that's being graphed is an integer, this returns true.
   * Otherwise, false.
   * @return Whether this object's data piece is an integer.
   *
   * <br>Edit history:
   * <br>------------------
   * March 29, 2005:  Created (LEM)
   */
  public boolean WantAnyGridInts() {
    return m_bIsInt;
  }

  /**
   * If our data piece that's being graphed is a float, this returns true.
   * Otherwise, false.
   * @return Whether this object's data piece is a float.
   *
   * <br>Edit history:
   * <br>------------------
   * March 29, 2005:  Created (LEM)
   */
  public boolean WantAnyGridFloats() {
    return m_bIsFloat;
  }

  /**
   * This nulls our dataset and recreates it, and clears all of our data
   * vectors.
   *
   * <br>Edit history:
   * <br>------------------
   * March 29, 2005:  Created (LEM)
   */
  void ClearData() {

    m_oDataset = null;
    m_oDataset = new ModelHistogramDataset();
    mp_oData.clear();
  }

  /**
   * Accepts the value of our data member from the parser, if float.
   * @param sGridName Name of the grid for this data
   * @param iX X number of the cell from which this value came
   * @param iY Y number of the cell from which this value came
   * @param iCode Data member code of this value.
   * @param fVal Value.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddGridFloatData(String sGridName, int iX, int iY, int iCode,
                               float fVal) {
    if (m_bIsFloat && sGridName.equals(m_sGridName) &&
        iCode == m_iDataMemberCode) {
      mp_oData.add(new Float(fVal));
    }
  }

  /**
   * This will add the integer data, if the data for this chart is an int that
   * matches the data member code.
   * @param sGridName Name of the grid for this data
   * @param iX X number of the cell from which this value came
   * @param iY Y number of the cell from which this value came
   * @param iCode Data member code of this value.
   * @param iVal Value.
   *
   * <br>Edit history:
   * <br>------------------
   * March 29, 2005:  Created (LEM)
   */
  public void AddGridIntData(String sGridName, int iX, int iY, int iCode,
                             int iVal) {
    if (m_bIsInt && sGridName.equals(m_sGridName) &&
        iCode == m_iDataMemberCode) {
      mp_oData.add(new Integer(iVal));
    }
  }
}
