package datavisualizer;

import javax.swing.*;
import org.jfree.data.DefaultCategoryDataset;

/**
* Creates different types of line graphs for detailed output files.  It can
* produce absolute and relative graphs of density, basal area, or volume for
* saplings, adults, and snags; and absolute and relative graphs of density
* alone for seedlings.
*
* This chart is different from other charts in that it displays data from all
* timesteps at once, instead of one timestep at a time.  Thus, it tends to
* ignore normal chart drawing requests and has its own code to force parsing
* of all timestep files in a detailed output package at once.
*
* Technically, density data could be extracted from any tree data member,
* since counting up the total is all that is required.  I didn't do it that
* way because I'm a little lazy, and I'll wait for a hue and cry from the
* users before unnecessarily complicating my code.
*
* Copyright: Copyright (c) Charles D. Canham 2003</p>
* Company: Institute of Ecosystem Studies</p>
* @author Lora E. Murphy
* @version 1.1
*
* <br>Edit history:
* <br>------------------
* <br>November 15, 2004:  Created (LEM)
* <br>November 18, 2004:  Updated for real-time data visualization (LEM)
* <br>March 9, 2005:  Added line graph support for volume and made version
* 1.1 (LEM)
*/
public class LineGraphDataRequest
    extends DataRequest {

  public final static int RELATIVE_BASAL_AREA = 1; /**<Chart type is relative basal area*/
  public final static int RELATIVE_DENSITY = 2; /**<Chart type is relative density*/
  public final static int ABSOLUTE_BASAL_AREA = 3; /**<Chart type is absolute basal area*/
  public final static int ABSOLUTE_DENSITY = 4; /**<Chart type is absolute density*/
  public final static int ABSOLUTE_VOLUME = 5; /**<Chart type is absolute volume*/
  public final static int RELATIVE_VOLUME = 6; /**<Chart type is relative volume*/

  /**The dataset with all species*/
  private DefaultCategoryDataset m_oDataset = null;

  /**Data array.  Array indexes are #1 - species, and #2 - timestep.*/
  private Float[][] mp_fData;

  /**Data totals for absolute chart types.  Array index is number of timestep.*/
  private Float[] mp_fTotal;

  /**Holds the data codes for the data being graphed.  For density for
   * seedlings, this is diam10; for density or basal area, it's DBH; for
   * volume, it's volume.  Array index is # of species.*/
  private int[] mp_iDataCode;

  /**A copy of which species are being displayed, so we know if the user
   * changed it*/
  private boolean[] mp_bWhichSpeciesShown;

  /**Area of the plot, in hectares*/
  private float m_fPlotAreaInHectares;

  /**The current timestep.  This helps process an entire detailed output file.
   * It also lets this class know when to ignore offered data - whenever this
   * value is set to -1, some other chart event triggered the parse.*/
  private int m_iCurrentTimestep = -1;

  /**The total number of timesteps.  We keep track of this so we know if
   * something has changed (as in real-time data visualization) so we can
   * update appropriately.*/
  private int m_iNumTimesteps = -1;

  /**The type of chart - absolute or relative / basal area or density.*/
  private int m_iChartType;

  /**Tree type of chart - seedling - snag.*/
  private int m_iTreeType;

  /**
   * Constructor.  Declares all the arrays.
   * @param oManager Parent detailed output file manager
   * @param sChartName Name of the table being drawn.
   * @param iChartType Type of graph, absolute or relative / density or basal
   * area.
   * @param iTreeType Type of tree for which to draw this chart.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>November 15, 2004:  Created (LEM)
   */
  public LineGraphDataRequest(DetailedOutputFileManager oManager,
                              String sChartName, int iChartType, int iTreeType) {
    super(sChartName, oManager);

    if (iChartType == ABSOLUTE_BASAL_AREA) {
      m_iChartType = ABSOLUTE_BASAL_AREA;
    }
    else if (iChartType == RELATIVE_BASAL_AREA) {
      m_iChartType = RELATIVE_BASAL_AREA;
    }
    else if (iChartType == RELATIVE_DENSITY) {
      m_iChartType = RELATIVE_DENSITY;
    }
    else if (iChartType == ABSOLUTE_VOLUME) {
      m_iChartType = ABSOLUTE_VOLUME;
    }
    else if (iChartType == RELATIVE_VOLUME) {
      m_iChartType = RELATIVE_VOLUME;
    }
    else {
      m_iChartType = ABSOLUTE_DENSITY;
    }

    if (iTreeType >= javawrapper.TreePopulation.SEEDLING &&
        iTreeType <= javawrapper.TreePopulation.SNAG) {
      m_iTreeType = iTreeType;
    }

    mp_iDataCode = new int[oManager.GetNumberOfSpecies()];
    int i;
    for (i = 0; i < mp_iDataCode.length; i++) {
      mp_iDataCode[i] = -1;
    }

    mp_bWhichSpeciesShown = new boolean[oManager.GetNumberOfSpecies()];
    for (i = 0; i < mp_bWhichSpeciesShown.length; i++) {
      mp_bWhichSpeciesShown[i] = false;
    }

    //Calculate the plot area
    float fPlotXLength = oManager.GetXPlotLength(),
        fPlotYLength = oManager.GetYPlotLength();

    m_fPlotAreaInHectares = fPlotXLength * fPlotYLength / 10000;

    //Get the number of timesteps
    m_iNumTimesteps = oManager.GetNumberOfActualTimesteps();

    //Say that this is a chart that doesn't show only one timestep at a time
    m_bShowOneTimestep = false;
  }

  /**
   * Writes the line graph's data to tab-delimited text.
   * @param jOut java.io.FileWriter The file to write to.
   * @throws java.io.IOException if there is a problem writing the file.
   * @throws ModelException passed through from called methods if the
   * dataset doesn't make sense.
   */
  protected void WriteChartDataToFile(java.io.FileWriter jOut) throws java.io.
      IOException, javawrapper.ModelException {

    int i, j;
    boolean bUseTotals = (mp_fTotal != null &&
                          (m_iChartType == ABSOLUTE_BASAL_AREA ||
                           m_iChartType == ABSOLUTE_DENSITY ||
                           m_iChartType == ABSOLUTE_VOLUME));

    //Write the chart header
    jOut.write(m_oChartFrame.getTitle() + "\n");
    if (m_iChartType == ABSOLUTE_BASAL_AREA) {
      jOut.write("Values in square meters of basal area per hectare.\n");
    }
    else if (m_iChartType == ABSOLUTE_DENSITY) {
      jOut.write("Values in number of trees per hectare.\n");
    }
    else if (m_iChartType == ABSOLUTE_VOLUME) {
      jOut.write("Values in cubic meters per hectare.\n");
    }

    //Make sure there's data
    if (mp_fData == null) {
      jOut.write("This chart has no data.");
      return;
    }

    //Write column headers
    Legend oLegend = m_oManager.GetLegend();
    jOut.write("Timestep");
    for (i = 0; i < mp_fData.length; i++) {
      jOut.write("\t" + oLegend.GetSpeciesDisplayName(i));
    }
    if (bUseTotals) {
      jOut.write("\tTotal");
    }
    jOut.write("\n");

    //Write data - #1 = species, #2 = timestep
    for (i = 0; i < mp_fData[0].length; i++) {
      jOut.write(String.valueOf(i));
      for (j = 0; j < mp_fData.length; j++) {
        jOut.write("\t" + mp_fData[j][i].toString());
      }
      if (bUseTotals) {
        jOut.write("\t" + mp_fTotal[i]);
      }
      jOut.write("\n");
    }
  }

  /**
   * This will update the chart only if the user has changed which species are
   * visible or unless the number of timesteps is different.
   * @param oLegend Legend The legend for this chart.
   * @throws ModelException Won't be thrown.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>November 15, 2004:  Created (LEM)
   * <br>November 18, 2004:  Updated for real-time data visualization (LEM)
   */
  void UpdateChart(Legend oLegend) throws javawrapper.ModelException {

    //Check to see if we need to respond for real-time data visualization
    boolean bRealTime = false;
    if ( ( (DetailedOutputLegend) oLegend).GetNumberOfTimesteps() != m_iNumTimesteps) {
      //Set this right away to make sure that, when this gets called again in
      //MakeDataset, we don't enter a loop
      m_iNumTimesteps = ( (DetailedOutputLegend) oLegend).GetNumberOfTimesteps();
      bRealTime = true;
    }

    //Check to see if the current set of visible species is different
    boolean[] p_bCurrentlyVisible = new boolean[oLegend.GetNumberOfSpecies()];
    int i;
    boolean bDifferent = false;

    //Get the currently visible species
    for (i = 0; i < p_bCurrentlyVisible.length; i++) {
      p_bCurrentlyVisible[i] = oLegend.GetIsSpeciesSelected(i);
    }

    //Check to see if they're the same
    for (i = 0; i < p_bCurrentlyVisible.length; i++) {
      if (p_bCurrentlyVisible[i] != mp_bWhichSpeciesShown[i]) {
        bDifferent = true;
        break;
      }
    }

    if (bRealTime == false && bDifferent == false) {
      return; //nothing's different
    }

    //If there's realtime visualization, make a new dataset
    if (bRealTime || m_oDataset == null) {
      m_oDataset = MakeDataset( (DetailedOutputLegend) oLegend);
    }

    //If the species list is different, save the current set of visible species
    if (bDifferent == true) {

      //Save the current set of visible species
      for (i = 0; i < p_bCurrentlyVisible.length; i++) {
        mp_bWhichSpeciesShown[i] = p_bCurrentlyVisible[i];
      }
    }

    //Draw the new chart
    DefaultCategoryDataset oAdjustedDataset = AdjustVisibleSpecies(m_oDataset,
        oLegend);

    DataGrapher.UpdateLineChart(oAdjustedDataset, m_oChartFrame, oLegend,
                                p_bCurrentlyVisible.length);
  }

  /**
   * Creates the chart dataset.
   * @param oLegend Legend The legend for this chart.
   * @return DefaultCategoryDataset The dataset for the chart, NOT adjusted for
   * which species are visible in the legend.
   * @throws ModelException if there is a problem creating the dataset.
   */
  private DefaultCategoryDataset MakeDataset(DetailedOutputLegend oLegend) throws
      javawrapper.ModelException {
    DetailedOutputFileManager oManager = oLegend.GetDetailedOutputFileManager();
    Float fVal;
    int iNumTimesteps = oLegend.GetNumberOfTimesteps(),
        iTimestepToReturnTo = oLegend.GetCurrentTimestep(),
        iNumSpecies = oLegend.GetNumberOfSpecies(),
        i, j;

    //Declare the data arrays
//    if (mp_fData == null) {
    mp_fData = new Float[iNumSpecies][iNumTimesteps + 1];
    mp_fTotal = new Float[iNumTimesteps + 1];
//    }

    //Force the processing of each timestep
    for (i = 0; i <= iNumTimesteps; i++) {

      //Stash the timestep we're on for the sake of AddTreeFloatData
      m_iCurrentTimestep = i;

      //Make the file manager parse this file
      oManager.ReadFile(i);
    }

    //Cause this chart to ignore data from other sources
    m_iCurrentTimestep = -1;

    //Refresh all the existing charts back to the way they were
    oManager.ReadFile(iTimestepToReturnTo);
    oManager.UpdateCharts();

    //Make sure there are no null values
    for (i = 0; i < mp_fData.length; i++) {
      for (j = 0; j < mp_fData[i].length; j++) {
        if (mp_fData[i][j] == null) {
          mp_fData[i][j] = new Float(0);
        }
      }
    }

    //Normalize all data to per-hectare values if absolute
    if (m_iChartType == ABSOLUTE_BASAL_AREA ||
        m_iChartType == ABSOLUTE_DENSITY ||
        m_iChartType == ABSOLUTE_VOLUME) {
      for (i = 0; i < mp_fData.length; i++) {
        for (j = 0; j < mp_fData[i].length; j++) {

          fVal = mp_fData[i][j];
          mp_fData[i][j] = new Float(fVal.floatValue() / m_fPlotAreaInHectares);
        }
      }
    }

    //Calculate totals
    for (i = 0; i < mp_fTotal.length; i++) {
      float fTotal = 0;
      for (j = 0; j < mp_fData.length; j++) {
        fTotal += mp_fData[j][i].floatValue();
      }
      mp_fTotal[i] = new Float(fTotal);
    }

    //If density, calculate proportions
    if (m_iChartType == RELATIVE_BASAL_AREA ||
        m_iChartType == RELATIVE_DENSITY ||
        m_iChartType == RELATIVE_VOLUME) {
      for (i = 0; i < mp_fTotal.length; i++) {
        if (mp_fTotal[i].floatValue() > 0) {
          for (j = 0; j < mp_fData.length; j++) {
            mp_fData[j][i] = new Float(mp_fData[j][i].floatValue() /
                                       mp_fTotal[i].floatValue());
          }
        }
      }
    }

    //Turn it into a dataset for the chart
    String[] p_sColumnKeys = new String[oLegend.GetNumberOfSpecies() + 1];

    //Create the row keys.  Rows equal timesteps.
    String[] p_sRowKeys = new String[mp_fData[0].length];
    for (i = 0; i < mp_fData[0].length; i++) {
      p_sRowKeys[i] = String.valueOf(i);
    }

    //Now the column keys.
    for (i = 0; i < p_sColumnKeys.length - 1; i++) {
      p_sColumnKeys[i] = oLegend.GetSpeciesCodeName(i);
    }

    //Last one for total, if applicable
    if (m_iChartType == ABSOLUTE_BASAL_AREA ||
        m_iChartType == ABSOLUTE_DENSITY ||
        m_iChartType == ABSOLUTE_VOLUME) {
      p_sColumnKeys[p_sColumnKeys.length - 1] = "Total";
    }

    // create the oDataset...
    DefaultCategoryDataset oDataset = new DefaultCategoryDataset();

    //Add each value, keyed to the proper row and column
    for (i = 0; i < p_sColumnKeys.length - 1; i++) {
      for (j = 0; j < mp_fData[i].length; j++) {
        oDataset.addValue(mp_fData[i][j], p_sColumnKeys[i], p_sRowKeys[j]);
      }
    }

    //Now add the total
    if (m_iChartType == ABSOLUTE_BASAL_AREA ||
        m_iChartType == ABSOLUTE_DENSITY ||
        m_iChartType == ABSOLUTE_VOLUME) {
      for (i = 0; i < mp_fTotal.length; i++) {
        oDataset.addValue(mp_fTotal[i], p_sColumnKeys[p_sColumnKeys.length - 1],
                          p_sRowKeys[i]);
      }
    }

    //Adjust the dataset to make sure only desired species are visible
    return oDataset;
  }

  /**
   * Draws the chart.
   * @param oLegend Legend Legend for this chart.
   * @param sChartTitle String Chart name for the window title.
   * @throws ModelException Passed through from other called methods - this
   * method doesn't throw it.
   * @return JInternalFrame The finished chart window.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>November 15, 2004:  Created (LEM)
   */
  ModelInternalFrame DrawChart(Legend oLegend, String sChartTitle) throws
      javawrapper.ModelException {

    //Get a copy of which species will be displayed
    for (int i = 0; i < oLegend.GetNumberOfSpecies(); i++) {
      mp_bWhichSpeciesShown[i] = oLegend.GetIsSpeciesSelected(i);
    }
    String sXAxisLabel = "Timesteps", sYAxisLabel = "";

    if (m_iChartType == ABSOLUTE_BASAL_AREA) {
      sYAxisLabel = "sq. m/ha";
    }
    else if (m_iChartType == RELATIVE_BASAL_AREA ||
             m_iChartType == RELATIVE_DENSITY ||
             m_iChartType == RELATIVE_VOLUME) {
      sYAxisLabel = "Proportion";
    }
    else if (m_iChartType == ABSOLUTE_DENSITY) {
      sYAxisLabel = "#/ha";
    }
    else if (m_iChartType == ABSOLUTE_VOLUME) {
      sYAxisLabel = "m3/ha";
    }
    else {
      return null;
    }

    m_oDataset = MakeDataset( (DetailedOutputLegend) oLegend);
    DefaultCategoryDataset oAdjustedDataset = AdjustVisibleSpecies(m_oDataset,
        oLegend);
    m_oChartFrame = DataGrapher.DrawLineChart(oAdjustedDataset, this,
                                              sXAxisLabel,
                                              sYAxisLabel,
                                              sChartTitle,
                                              oLegend,
                                              oLegend.GetNumberOfSpecies());
    return m_oChartFrame;
  }

  /**
   * Does nothing, because the charts don't operate on the detailed output
   * file update protocol.
   * @throws ModelException Won't throw it.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>November 15, 2004:  Created (LEM)
   */
  void ClearData() throws javawrapper.ModelException {
    ;
  }

  /**
   * This wants diameter values for all trees.
   * @return boolean Set to true.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>November 15, 2004:  Created (LEM)
   */
  public boolean WantAnyTreeFloats() {
    return true;
  }

  /**
   * Accepts a tree float data member code.  We're looking for DBH and diam10.
   * @param iSpecies The species for which this is a data member.
   * @param iType The tree type for which this is a data member.
   * @param sLabel The label of the data member.
   * @param iCode The data member code.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>November 15, 2004:  Created (LEM)
   * <br>March 9, 2005:  Updated to look for volume (LEM)
   */
  public void AddTreeFloatDataMemberCode(int iSpecies, int iType, String sLabel,
                                         int iCode) {

    if (iType == m_iTreeType) {
      //03/09/05 LEM - check for the appropriate chart type
      if ((m_iChartType == ABSOLUTE_VOLUME ||
          m_iChartType == RELATIVE_VOLUME) &&
         sLabel.equalsIgnoreCase("volume")) {
       mp_iDataCode[iSpecies] = iCode;
     }
     else if (sLabel.equalsIgnoreCase("dbh")) {
        mp_iDataCode[iSpecies] = iCode;
      }
      else if (sLabel.equalsIgnoreCase("diam10") &&
               javawrapper.TreePopulation.SEEDLING == iType) {
        mp_iDataCode[iSpecies] = iCode;
      }
    }
  }

  /**
   * Accepts a tree float data member value.  If it matches our code, we'll
   * increment either the density count or the basal area total, depending
   * on the chart type.
   * @param iSpecies Species of the tree from which this value came.
   * @param iType Type of the tree from which this value came.
   * @param iCode Data member code of this value.
   * @param fVal Value.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>November 15, 2004:  Created (LEM)
   * <br>March 9, 2005:  Added volume (LEM)
   */
  public void AddTreeFloatData(int iSpecies, int iType, int iCode, float fVal) {

    //If this was triggered by some other event, ignore
    if ( -1 == m_iCurrentTimestep) {
      return;
    }

    //Make sure it matches our type
    if (iType != m_iTreeType) {
      return;
    }

    if (iCode == mp_iDataCode[iSpecies]) {

      Float fCurrVal = (Float) mp_fData[iSpecies][m_iCurrentTimestep];
      if (fCurrVal == null) {
        fCurrVal = new Float(0);
      }

      if (m_iChartType == ABSOLUTE_DENSITY || m_iChartType == RELATIVE_DENSITY) {
        //Increment the density count
        fCurrVal = new Float(fCurrVal.floatValue() + 1);
      }
      else if (m_iChartType == ABSOLUTE_VOLUME || m_iChartType == RELATIVE_VOLUME) {
        //03/09/05 LEM - increment volume
        fCurrVal = new Float(fCurrVal.floatValue() + fVal);
      }
      else {

        //Increment the basal area
        fCurrVal = new Float(fCurrVal.floatValue() +
                             java.lang.Math.pow(fVal * 0.5, 2) *
                             java.lang.Math.PI * 0.0001);
      }

      //Set the value back into the table
      mp_fData[iSpecies][m_iCurrentTimestep] = fCurrVal;
    }
  }

  /**
   * This will take a dataset and make a copy with only series for species
   * which are supposed to be visible.
   * @param oDataset The dataset.
   * @param oLegend The legend which controls which species are visible.
   * @return The dataset copy.
   * @throws ModelException wrapping another exception.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>November 15, 2004: Created (LEM)
   */
  protected DefaultCategoryDataset AdjustVisibleSpecies(DefaultCategoryDataset
      oDataset, Legend oLegend) throws javawrapper.ModelException {

    Comparable oRowKey, oColKey;
    int iSp, iNumSpecies = oLegend.GetNumberOfSpecies(), iRow, iCol, iNumRows,
        iNumCols, i;

    //Make a copy of the new dataset - cloning doesn't work
    DefaultCategoryDataset oNewDataset = new DefaultCategoryDataset();
    iNumRows = oDataset.getRowCount();
    iNumCols = oDataset.getColumnCount();
    for (iCol = 0; iCol < iNumCols; iCol++) {
      oColKey = oDataset.getColumnKey(iCol);
      for (iRow = 0; iRow < iNumRows; iRow++) {
        oRowKey = oDataset.getRowKey(iRow);
        oNewDataset.setValue(oDataset.getValue(oRowKey, oColKey), oRowKey,
                             oColKey);
      }
    }

    for (iSp = 0; iSp < iNumSpecies; iSp++) {
      //The "rows" are what contain the species
      if (!oLegend.GetIsSpeciesSelected(iSp)) {
        for (i = 0; i < oNewDataset.getRowCount(); i++) {
          if (oNewDataset.getRowKey(i).compareTo(oLegend.GetSpeciesCodeName(iSp)) ==
              0) {
            oNewDataset.removeRow(i);
          }
        }
      }
    }
    return oNewDataset;
  }

}
