package datavisualizer;

import javax.swing.*;

/**
 * Writes an overview table for a detailed output file (basal area and density
 * for each life history stage).  The tabulation of density and basal area data
 * is based on the diameter value for each tree life history stage (that is,
 * diameter at 10 cm for seedlings and DBH for everything else).  For density,
 * the number of diameter values encountered is counted; for basal area, the
 * individual basal areas calculated from DBH are totaled.  All values are
 * displayed in per hectare units.
 * <p>
 * Technically, density data could be extracted from any tree data member,
 * since counting up the total is all that is required.  I didn't do it that
 * way because I'm a little lazy, and I'll wait for a hue and cry from the
 * users before unnecessarily complicating my code.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>August 4, 2004:  Created (LEM)
 */

public class OverviewTableDataRequest
    extends DataRequest
    implements java.awt.event.ActionListener {

  /**Basal area data.  Array indexes are #1 - type and #2 - species.*/
  private float[][] mp_fABA;

  /**Absolute density data.  Array indexes are #1 - type and #2 - species.*/
  private int[][] mp_iADN;

  /**Holds the data codes for diameter (diam10 or DBH).  Array indexes are
   * #1 - type and #2 - species.*/
  private int[][] mp_iDbhCode;

  /**Area of the plot, in hectares*/
  private float m_fPlotAreaInHectares;

  /**Number of total species*/
  private int m_iNumSpecies;

  /**
   * Constructor.  Declares all the arrays.
   * @param oManager Parent detailed output file manager
   * @param sChartName Name of the table being drawn.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>August 4, 2004:  Created (LEM)
   */
  public OverviewTableDataRequest(DetailedOutputFileManager oManager,
                                  String sChartName) {
    super(sChartName, oManager);

    m_iNumSpecies = oManager.GetNumberOfSpecies();
    int iNumTypes = oManager.GetNumberOfTypes();

    mp_fABA = new float[iNumTypes][m_iNumSpecies];
    mp_iADN = new int[iNumTypes][m_iNumSpecies];
    mp_iDbhCode = new int[iNumTypes][m_iNumSpecies];

    //Calculate the plot area
    float fPlotXLength = oManager.GetXPlotLength(),
        fPlotYLength = oManager.GetYPlotLength();

    m_fPlotAreaInHectares = fPlotXLength * fPlotYLength / 10000;

    //For now - set that this can't write the whole run at once.  This requires
    //special handling to do it right.  Right now the data comes from the
    //tables but they wouldn't be updated in an all-timesteps situation -
    //we have to change this to write from the data
    m_bShowOneTimestep = false;
  }

  /**
   * Causes the table to refresh its data.
   * @param oLegend Legend The legend for this chart.
   * @throws ModelException Won't be thrown.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>August 4, 2004:  Created (LEM)
   */
  void UpdateChart(Legend oLegend) throws javawrapper.ModelException {
    //Pass the datasets to create the table
    Object[][][] p_oAllDatasets = CreateDataset();

    DataGrapher.UpdateOverviewTables( (javax.swing.JPanel) m_oChartFrame.
                                     getContentPane(), p_oAllDatasets[0],
                                     p_oAllDatasets[1], p_oAllDatasets[2],
                                     p_oAllDatasets[3]);
  }

  /**
   * Creates a table dataset out of the accumulated data.  The four
   * primary arrays within the overall array are ready to pass to the
   * table constructor.
   * @return Object[][][] Array of datasets for tables, ready to pass to
   * DataGrapher. Index 0 is the seedling dataset, 1 is the sapling dataset, 2
   * is the adult dataset, and 3 is the snag dataset.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>August 4, 2004:  Created (LEM)
   */
  private Object[][][] CreateDataset() {
    Object[][][] p_oAllDatasets = new Object[4][][];

    java.text.NumberFormat jFormat = java.text.NumberFormat.getInstance();
    jFormat.setMaximumFractionDigits(3);
    jFormat.setMinimumFractionDigits(3);

    java.text.NumberFormat jRelFormat = java.text.NumberFormat.getPercentInstance();
    jRelFormat.setMaximumFractionDigits(2);
    jRelFormat.setMinimumFractionDigits(2);

    //Set up the datasets
    float fTotalDensity, fTotalBasalArea;
    int i;

    //*****************************
     // Adults table dataset
     //*****************************
      p_oAllDatasets[2] = new Object[m_iNumSpecies + 1][4]; //rows by columns

      //Total up basal area and density
    fTotalDensity = 0;
    fTotalBasalArea = 0;
    for (i = 0; i < m_iNumSpecies; i++) {
      fTotalDensity += mp_iADN[javawrapper.TreePopulation.ADULT][i];
      fTotalBasalArea += mp_fABA[javawrapper.TreePopulation.ADULT][i];
    }

    //Create a row for each species
    for (i = 0; i < m_iNumSpecies; i++) {

      //Column 1: absolute density - normalize to per hectare
      p_oAllDatasets[2][i][0] = jFormat.format(mp_iADN[javawrapper.
                                               TreePopulation.
                                               ADULT][i] /
                                               m_fPlotAreaInHectares);

      //Column 2: percent of density
      if (fTotalDensity > 0) {
        p_oAllDatasets[2][i][1] = jRelFormat.format(mp_iADN[javawrapper.
                                                 TreePopulation.ADULT][i] /
                                                 fTotalDensity);
      }
      else {
        p_oAllDatasets[2][i][1] = new Float(0);
      }

      //Column 3: absolute basal area
      p_oAllDatasets[2][i][2] = jFormat.format(mp_fABA[javawrapper.
                                               TreePopulation.
                                               ADULT][i] /
                                               m_fPlotAreaInHectares);

      //Column 4: percent of basal area
      if (fTotalBasalArea > 0) {
        p_oAllDatasets[2][i][3] = jRelFormat.format(mp_fABA[javawrapper.
                                                 TreePopulation.ADULT][i] /
                                                 fTotalBasalArea);
      }
      else {
        p_oAllDatasets[2][i][3] = new Float(0);
      }
    }

    //Row for totals
    //Column 1: absolute density
    p_oAllDatasets[2][m_iNumSpecies][0] = jFormat.format(fTotalDensity /
        m_fPlotAreaInHectares);

    //Column 3: absolute basal area
    p_oAllDatasets[2][m_iNumSpecies][2] = jFormat.format(fTotalBasalArea /
        m_fPlotAreaInHectares);

    //*****************************
     // Saplings table dataset
     //*****************************
    p_oAllDatasets[1] = new Object[m_iNumSpecies + 1][4]; //rows by columns

    //Total up basal area and density
    fTotalDensity = 0;
    fTotalBasalArea = 0;
    for (i = 0; i < m_iNumSpecies; i++) {
      fTotalDensity += mp_iADN[javawrapper.TreePopulation.SAPLING][i];
      fTotalBasalArea += mp_fABA[javawrapper.TreePopulation.SAPLING][i];
    }

    //Create a row for each species
    for (i = 0; i < m_iNumSpecies; i++) {

      //Column 1: absolute density - normalize to per hectare
      p_oAllDatasets[1][i][0] = jFormat.format(mp_iADN[javawrapper.
                                               TreePopulation.
                                               SAPLING][i] /
                                               m_fPlotAreaInHectares);

      //Column 2: percent of density
      if (fTotalDensity > 0) {
        p_oAllDatasets[1][i][1] = jRelFormat.format(mp_iADN[javawrapper.
                                                 TreePopulation.SAPLING][i] /
                                                 fTotalDensity);
      }
      else {
        p_oAllDatasets[1][i][1] = new Float(0);
      }

      //Column 3: absolute basal area
      p_oAllDatasets[1][i][2] = jFormat.format(mp_fABA[javawrapper.
                                               TreePopulation.
                                               SAPLING][i] /
                                               m_fPlotAreaInHectares);

      //Column 4: percent of basal area
      if (fTotalBasalArea > 0) {
        p_oAllDatasets[1][i][3] = jRelFormat.format(mp_fABA[javawrapper.
                                                 TreePopulation.SAPLING][i] /
                                                 fTotalBasalArea);
      }
      else {
        p_oAllDatasets[1][i][3] = new Float(0);
      }
    }

    //Row for totals
    //Column 1: absolute density
    p_oAllDatasets[1][m_iNumSpecies][0] = jFormat.format(fTotalDensity /
        m_fPlotAreaInHectares);

    //Column 3: absolute basal area
    p_oAllDatasets[1][m_iNumSpecies][2] = jFormat.format(fTotalBasalArea /
        m_fPlotAreaInHectares);

    //*****************************
     // Snags table dataset
     //*****************************
    p_oAllDatasets[3] = new Object[m_iNumSpecies + 1][4]; //rows by columns

    //Total up basal area and density
    fTotalDensity = 0;
    fTotalBasalArea = 0;
    for (i = 0; i < m_iNumSpecies; i++) {
      fTotalDensity += mp_iADN[javawrapper.TreePopulation.SNAG][i];
      fTotalBasalArea += mp_fABA[javawrapper.TreePopulation.SNAG][i];
    }

    //Create a row for each species
    for (i = 0; i < m_iNumSpecies; i++) {

      //Column 1: absolute density - normalize to per hectare
      p_oAllDatasets[3][i][0] = jFormat.format(mp_iADN[javawrapper.
                                               TreePopulation.
                                               SNAG][i] / m_fPlotAreaInHectares);

      //Column 2: percent of density
      if (fTotalDensity > 0) {
        p_oAllDatasets[3][i][1] = jRelFormat.format(mp_iADN[javawrapper.
                                                 TreePopulation.SNAG][i] /
                                                 fTotalDensity);
      }
      else {
        p_oAllDatasets[3][i][1] = new Float(0);
      }

      //Column 3: absolute basal area
      p_oAllDatasets[3][i][2] = jFormat.format(mp_fABA[javawrapper.
                                               TreePopulation.
                                               SNAG][i] / m_fPlotAreaInHectares);

      //Column 4: percent of basal area
      if (fTotalBasalArea > 0) {
        p_oAllDatasets[3][i][3] = jRelFormat.format(mp_fABA[javawrapper.
                                                 TreePopulation.SNAG][i] /
                                                 fTotalBasalArea);
      }
      else {
        p_oAllDatasets[3][i][3] = new Float(0);
      }
    }

    //Row for totals
    //Column 1: absolute density
    p_oAllDatasets[3][m_iNumSpecies][0] = jFormat.format(fTotalDensity /
        m_fPlotAreaInHectares);

    //Column 3: absolute basal area
    p_oAllDatasets[3][m_iNumSpecies][2] = jFormat.format(fTotalBasalArea /
        m_fPlotAreaInHectares);

    //*****************************
     // Seedlings table dataset
     //*****************************
      //Create our table data array
    p_oAllDatasets[0] = new Object[m_iNumSpecies + 1][1]; //rows by columns

    //Total up density
    fTotalDensity = 0;
    for (i = 0; i < m_iNumSpecies; i++) {
      fTotalDensity += mp_iADN[javawrapper.TreePopulation.SEEDLING][i];
    }

    //Create a row for each species
    for (i = 0; i < m_iNumSpecies; i++) {

      //Column 1: absolute density
      p_oAllDatasets[0][i][0] = jFormat.format(mp_iADN[javawrapper.
                                               TreePopulation.
                                               SEEDLING][i] /
                                               m_fPlotAreaInHectares);
    }

    //Row for totals
    //Column 1: absolute density
    p_oAllDatasets[0][m_iNumSpecies][0] = jFormat.format(fTotalDensity /
        m_fPlotAreaInHectares);

    return p_oAllDatasets;
  }

  /**
   * Draws the table.
   * @param oLegend Legend Legend for this chart.
   * @param sChartTitle String Chart name for the window title.
   * @throws ModelException Passed through from other called methods - this
   * method doesn't throw it.
   * @return JInternalFrame The finished chart window.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>August 4, 2004:  Created (LEM)
   */
  ModelInternalFrame DrawChart(Legend oLegend, String sChartTitle) throws
      javawrapper.ModelException {

    //Pass the datasets to create the table
    Object[][][] p_oAllDatasets = CreateDataset();
    ModelInternalFrame jFrame = DataGrapher.DrawOverviewTables(
        sChartTitle, p_oAllDatasets[0], p_oAllDatasets[1], p_oAllDatasets[2],
        p_oAllDatasets[3], oLegend, this);

    m_oChartFrame = jFrame;
    return jFrame;
  }

  /**
   * Sets all the values in the data collection arrays back to 0.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>August 4, 2004:  Created (LEM)
   */
  void ClearData() {
    int i, j;

    for (i = 0; i < mp_fABA.length; i++) {
      for (j = 0; j < mp_fABA[i].length; j++) {
        mp_fABA[i][j] = 0;
        mp_iADN[i][j] = 0;
      }
    }
  }

  /**
   * This wants diameter values for all trees.
   * @return boolean Set to true.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>August 4, 2004:  Created (LEM)
   */
  public boolean WantAnyTreeFloats() {
    return true;
  }

  /**
   * Accepts a tree float data member code.  We're looking for diam10 for
   * seedlings, and DBH for all other tree types supported.
   * @param iSpecies The species for which this is a data member.
   * @param iType The tree type for which this is a data member.
   * @param sLabel The label of the data member.
   * @param iCode The data member code.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>August 4, 2004:  Created (LEM)
   */
  public void AddTreeFloatDataMemberCode(int iSpecies, int iType, String sLabel,
                                         int iCode) {

    if (sLabel.equalsIgnoreCase("dbh")) {
      mp_iDbhCode[iType][iSpecies] = iCode;
    }
    else if (sLabel.equalsIgnoreCase("diam10") &&
             iType == javawrapper.TreePopulation.SEEDLING) {
      mp_iDbhCode[iType][iSpecies] = iCode;
    }
  }

  /**
   * Accepts a tree float data member value.  If it matches our code, we'll
   * increment the density count and the basal area total, if it's bigger
   * than a seedling.
   * @param iSpecies Species of the tree from which this value came.
   * @param iType Type of the tree from which this value came.
   * @param iCode Data member code of this value.
   * @param fVal Value.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>August 4, 2004:  Created (LEM)
   */
  public void AddTreeFloatData(int iSpecies, int iType, int iCode, float fVal) {

    if (iType < javawrapper.TreePopulation.SEEDLING &&
        iType > javawrapper.TreePopulation.SNAG) {
      return;
    }

    if (iCode == mp_iDbhCode[iType][iSpecies]) {

      //Increment the density count
      mp_iADN[iType][iSpecies]++;

      //Increment the basal area
      if (iType > javawrapper.TreePopulation.SEEDLING) {
        mp_fABA[iType][iSpecies] += java.lang.Math.pow(fVal * 0.5, 2) *
            java.lang.Math.PI * 0.0001;
      }
    }
  }

  /**
   * Saves the data in the current table as a tab-delimited text file.
   * @param jOut java.io.FileWriter The file to write to.
   * @throws java.io.IOException if there's a problem writing the file.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>November 12, 2004: Created (LEM)
   */
  protected void WriteChartDataToFile(java.io.FileWriter jOut) throws java.io.
      IOException {

    String[] p_sNames = new String[] {
        "seedling_table", "sapling_table",
        "adult_table", "snag_table"};

    int i, j;

    //Write timestep
    DetailedOutputLegend oLegend = (DetailedOutputLegend) m_oManager.GetLegend();
    jOut.write("Timestep\t" + String.valueOf(oLegend.GetCurrentTimestep()) + "\n");

    //Find each table
    java.awt.Component jContentPane = m_oChartFrame.getContentPane();
    for (i = 0; i < 4; i++) {
      JPanel jTablePanel = (JPanel) DataGrapher.FindNamedComponent(
          jContentPane, "table_panel_" + i);

      //First find the label heading the panel
      for (j = 0; j < jTablePanel.getComponentCount(); j++) {
        if (jTablePanel.getComponent(j) instanceof JLabel) {
          jOut.write("\n" +
                     ( (JLabel) jTablePanel.getComponent(j)).getText() +
                     "\n");
          break;
        }
      }

      //Now find the table
      JTable jTable = (JTable) DataGrapher.FindNamedComponent(jTablePanel,
          p_sNames[i]);
      int iNumRows = jTable.getRowCount();
      int iNumCols = jTable.getColumnCount();
      int iRow, iCol;

      //Write header row
      jOut.write(String.valueOf(jTable.getTableHeader().getColumnModel().
                                getColumn(0).getHeaderValue()));
      for (iCol = 1; iCol < iNumCols; iCol++) {
        jOut.write("\t" +
                   String.valueOf(jTable.getTableHeader().getColumnModel().
                                  getColumn(iCol).getHeaderValue()));
      }
      jOut.write("\n");

      for (iRow = 0; iRow < iNumRows; iRow++) {
        jOut.write(String.valueOf(jTable.getValueAt(iRow, 0)));
        for (iCol = 1; iCol < iNumCols; iCol++) {
          jOut.write("\t" +
                     String.valueOf(jTable.getValueAt(iRow, iCol)));
        }
        jOut.write("\n");
      }
    }
  }
}
