package datavisualizer;

import org.jfree.data.*;
import java.util.List;
import java.util.Collections;
import org.jfree.util.ObjectUtils;

/**
 * For creating an XYZ dataset.  I, Lora E. Murphy, am only the author in the
 * weakest possible sense.  I
 * got this wholesale from the JFreeChart package - XYSeriesCollection.java.
 * I merely added the method for getZValue so that this implements the
 * XYZDataset interface.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 */

public class XYZSeriesCollection
    extends AbstractSeriesDataset
    implements XYZDataset {
  /** The series that are included in the collection. */
  private List data;

  /**
   * Constructs an empty dataset.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public XYZSeriesCollection() {

    this.data = new java.util.ArrayList();

  }

  /**
   * Constructs a dataset and populates it with a single time series.
   *
   * @param series  the time series.
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public XYZSeriesCollection(XYZSeries series) {

    this.data = new java.util.ArrayList();
    if (series != null) {
      data.add(series);
      series.addChangeListener(this);
    }

  }

  /**
   * Adds a series to the collection.
   * <P>
   * Notifies all registered listeners that the dataset has changed.
   *
   * @param series  the series.
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void addSeries(XYZSeries series) {

    // check arguments...
    if (series == null) {
      throw new IllegalArgumentException(
          "XYZSeriesCollection.addSeries(...): cannot add null series.");
    }

    // add the series...
    data.add(series);
    series.addChangeListener(this);
    fireDatasetChanged();

  }

  /**
   * Returns the number of series in the collection.
   *
   * @return the number of series in the collection.
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public int getSeriesCount() {
    return this.data.size();
  }

  /**
   * Returns a list of all the series in the collection.
   *
   * @return The list (which is unmodifiable).
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public List getSeries() {
    return Collections.unmodifiableList(this.data);
  }

  /**
   * Returns a series.
   *
   * @param series  the series (zero-based index).
   *
   * @return The series.
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public XYZSeries getSeries(int series) {

    // check arguments...
    if ( (series < 0) || (series > getSeriesCount())) {
      throw new IllegalArgumentException(
          "XYZSeriesCollection.getSeries(...): index outside valid range.");
    }

    // fetch the series...
    XYZSeries ts = (XYZSeries) data.get(series);
    return ts;

  }

  /**
   * Returns the name of a series.
   *
   * @param series  the series (zero-based index).
   *
   * @return the name of a series.
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public String getSeriesName(int series) {

    // check arguments...delegated

    // fetch the result...
    return getSeries(series).getName();

  }

  /**
   * Returns the number of items in the specified series.
   *
   * @param series  the series (zero-based index).
   *
   * @return the number of items in the specified series.
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public int getItemCount(int series) {

    // check arguments...delegated

    // fetch the result...
    return getSeries(series).getItemCount();

  }

  /**
   * Returns the x-value for the specified series and item.
   *
   * @param series  the series (zero-based index).
   * @param item  the item (zero-based index).
   *
   * @return the x-value for the specified series and item.
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public Number getXValue(int series, int item) {

    XYZSeries ts = (XYZSeries) data.get(series);
    XYZDataItem XYZItem = ts.getDataItem(item);
    return XYZItem.getX();

  }

  /**
   * Returns the y-value for the specified series and item.
   *
   * @param series  the series (zero-based index).
   * @param index  the index of the item of interest (zero-based).
   *
   * @return the y-value for the specified series and item.
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public Number getYValue(int series, int index) {

    XYZSeries ts = (XYZSeries) data.get(series);
    XYZDataItem XYZItem = ts.getDataItem(index);
    return XYZItem.getY();

  }

  /**
   * Removes all the series from the collection.
   * <P>
   * Notifies all registered listeners that the dataset has changed.
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void removeAllSeries() {

    // Unregister the collection as a change listener to each series in the collection.
    for (int i = 0; i < this.data.size(); i++) {
      XYZSeries series = (XYZSeries)this.data.get(i);
      series.removeChangeListener(this);
    }

    // Remove all the series from the collection and notify listeners.
    data.clear();
    fireDatasetChanged();

  }

  /**
   * Removes a series from the collection.
   * <P>
   * Notifies all registered listeners that the dataset has changed.
   *
   * @param series  the series.
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void removeSeries(XYZSeries series) {

    // check arguments...
    if (series == null) {
      throw new IllegalArgumentException(
          "XYZSeriesCollection.removeSeries(...): cannot remove null series.");
    }

    // remove the series...
    if (data.contains(series)) {
      series.removeChangeListener(this);
      data.remove(series);
      fireDatasetChanged();
    }

  }

  /**
   * Removes a series from the collection.
   * <P>
   * Notifies all registered listeners that the dataset has changed.
   *
   * @param series  the series (zero based index).
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void removeSeries(int series) {

    // check arguments...
    if ( (series < 0) || (series > getSeriesCount())) {
      throw new IllegalArgumentException(
          "XYZSeriesCollection.removeSeries(...): index outside valid range.");
    }

    // fetch the series, remove the change listener, then remove the series.
    XYZSeries ts = (XYZSeries) data.get(series);
    ts.removeChangeListener(this);
    data.remove(series);
    fireDatasetChanged();

  }

  /**
   * Tests this collection for equality with an arbitrary object.
   *
   * @param obj  the object.
   *
   * @return A boolean.
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean equals(Object obj) {

    if (obj == null) {
      return false;
    }

    if (obj == this) {
      return true;
    }

    if (obj instanceof XYZSeriesCollection) {
      XYZSeriesCollection c = (XYZSeriesCollection) obj;
      return ObjectUtils.equal(this.data, c.data);
    }

    return false;

  }

  /**
   * Returns the z-value for the specified series and item.
   *
   * @param series  the series (zero-based index).
   * @param index  the index of the item of interest (zero-based).
   *
   * @return the z-value for the specified series and item.
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public Number getZValue(int series, int index) {

    XYZSeries ts = (XYZSeries) data.get(series);
    XYZDataItem xyzItem = ts.getDataItem(index);
    return xyzItem.getZ();

  }

}