package javawrapper;

import java.io.*;
import java.util.Vector;
import org.xml.sax.Attributes;

/**
 * Manages planting behaviors and data.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 */

public class PlantingBehaviors
    extends BehaviorTypeBase {

  /**Gridded planting.  Matches old code enum value - important!*/
  public static final int GRIDDED = 1;
  /**Random planting.  Matches old code enum value - important!*/
  public static final int RANDOM = 0;

  /**Our array of plantings*/
  protected Vector mp_oPlantings;

  /**List of diam10 values for newly planted trees - one for each species*/
  /**Slope of growth response for each species*/
  protected ModelVector mp_fInitialDiam10 = new ModelVector(
      "Avg. diameter at 10 cm for new seedlings",
      "pl_initialDiam10",
      "pl_idVal", 0,
      ModelVector.FLOAT);

  /**
   * Constructor.
   * @param oManager GUIManager object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public PlantingBehaviors(GUIManager oManager) {
    super(oManager, "Planting", "planting");
    m_bAutoDisplayParameters = false;

    mp_oPlantings = new Vector(0);

    //Set up planting behaviors
    mp_oChildBehaviors = new Behavior[1];

    //planting behavior
    mp_oChildBehaviors[0] = new Behavior("Planting", "Planting", "plant", (float)1);
    mp_oChildBehaviors[0].m_bCanEditAutomatically = false;
    mp_oChildBehaviors[0].m_bMustHaveTrees = false;
    mp_oChildBehaviors[0].AddRequiredData(mp_fInitialDiam10);

    //Put our initial diam10 data in the required data list
    mp_oAllData.add(mp_fInitialDiam10);
  }

  /**
   * Reads in an old tab-delimited harvest regime file.  Any existing planting
   * data will be replaced with what's in the file.
   * @param sFileName Filename of harvest regime file.
   * @param oPlot Plot object.
   * @param oPop TreePopulation object.
   * @throws ModelException Wrapping an IO exception, or if the harvest
   * parameters do not match the currently loaded parameter file.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void ReadOldHarvestRegimeFile(String sFileName, Plot oPlot,
                                       TreePopulation oPop) throws
      ModelException {
    FileReader oIn = null;
    try {
      oIn = new FileReader(sFileName);
      Cell[] p_oCells;
      PlantingData oPlanting = null;
      Vector oLine, oPlantingList = new Vector(0);
      String sVal;
      Integer iVal;
      float fAbundance,
          fOldXCellLength, fOldYCellLength; //old X and Y cell lengths
      int iNumOldGridCells,
          iOldNumXGrids, iOldNumYGrids, //old number of X and Y grids
          iNumSpecies = oPop.GetNumberOfSpecies(),
          iTimestep,
          iLoadFlag,
          iX, iY,
          iGrid, i;
      ModelFileFunctions.SkipLine(oIn);
      //Get the first data line
      oLine = ModelFileFunctions.ReadLine(oIn);
      //Validate the plot lengths - remember they're reversed from the old to the
      //new coordinate system
      sVal = (String) oLine.elementAt(2); //old X cell length
      fOldXCellLength = new Float(sVal).floatValue();
      sVal = (String) oLine.elementAt(0); //old number of X cells
      iOldNumXGrids = new Integer(sVal).intValue();
      sVal = (String) oLine.elementAt(3); //old Y plot length
      fOldYCellLength = new Float(sVal).floatValue();
      sVal = (String) oLine.elementAt(1); //old number of Y cells
      iOldNumYGrids = new Integer(sVal).intValue();
      if (fOldXCellLength * (float) iOldNumXGrids != oPlot.GetPlotYLength() ||
          fOldYCellLength * (float) iOldNumYGrids != oPlot.GetPlotXLength()) {
        ModelException oErr = new ModelException();
        oErr.sCallingFunction = "JAVA";
        oErr.iErrorCode = ErrorGUI.BAD_FILE;
        oErr.sMessage = "The plot dimensions for this harvest regime file " +
            "do not match the plot dimensions for this parameter file.";
        throw oErr;
      }
      //Validate the number of species
      sVal = (String) oLine.elementAt(4);
      if (new Integer(sVal).intValue() != iNumSpecies) {
        ModelException oErr = new ModelException();
        oErr.sCallingFunction = "JAVA";
        oErr.iErrorCode = ErrorGUI.BAD_FILE;
        oErr.sMessage = "The number of species for this harvest regime file (" +
            sVal +
            ") does not match the plot length for this parameter file (" +
            String.valueOf(iNumSpecies) + ".";
        throw oErr;
      }
      //Check the what's to load flag - if harvest only, skip out
      sVal = (String) oLine.elementAt(5);
      iLoadFlag = new Integer(sVal).intValue();
      if (iLoadFlag == 1) { //1 = harvest only
        return;
      }
      //Calculate the number of grid cells in this file
      sVal = (String) oLine.elementAt(0); //number of X cells
      iNumOldGridCells = new Integer(sVal).intValue();
      sVal = (String) oLine.elementAt(1); //number of Y cells
      iNumOldGridCells *= new Integer(sVal).intValue();

      //If there is both planting and harvest data in this file, we need
      //to skip over the harvest stuff, which is first
      if (iLoadFlag == 0) { //load both harvest and planting
        for (iGrid = 0; iGrid < iNumOldGridCells; iGrid++) {
          for (i = 0; i < iNumSpecies; i++) {
            //Read the lines for this species and grid cell until there is a
            //end of record flag (-1) in the first column
            oLine = ModelFileFunctions.ReadLine(oIn);
            sVal = (String) oLine.elementAt(0);
            iVal = new Integer(sVal);
            while (iVal.intValue() > -1) {
              oLine = ModelFileFunctions.ReadLine(oIn);
              sVal = (String) oLine.elementAt(0);
              iVal = new Integer(sVal);
            }
          }
        }
      }

      //Now we've advanced to the planting data - so read it
      for (iGrid = 0; iGrid < iNumOldGridCells; iGrid++) {

        //Then create event records and read in the data
        oLine = ModelFileFunctions.ReadLine(oIn);
        //When this record is over, it will be designated by a -1 in the
        //first bucket of the line.  Read all the info for this grid and
        //species until we get to that end-of-record designator.
        sVal = (String) oLine.elementAt(0);
        iVal = new Integer(sVal);
        while (iVal.intValue() > -1) {

          //If the line only has one value - it's got no data
          if (oLine.size() > 1) {

            //First value is grid cell - make sure it matches our current grid
            if (iVal.intValue() != iGrid) {
              ModelException oErr = new ModelException();
              oErr.iErrorCode = ErrorGUI.TAB_FILE_PREMATURE_END;
              oErr.sCallingFunction = "ReadOldHarvestRegimeFile";
              oErr.sMessage = sFileName;
            }
            //Second value is timestep
            sVal = (String) oLine.elementAt(1);
            iTimestep = new Integer(sVal).intValue();
            if (iTimestep > -1) {
              //Create a new planting data record
              oPlanting = null;
              oPlanting = new PlantingData();
              //Set timestep
              oPlanting.SetTimestep(iTimestep);

              //The next several values are plant abundances - add each
              //species with a positive abundance to the list
              for (i = 0; i < iNumSpecies; i++) {
                sVal = (String) oLine.elementAt(i + 2);
                fAbundance = new Float(sVal).floatValue();
                if (fAbundance > 0) {
                  oPlanting.AddAbundance(i, fAbundance);
                  oPlanting.AddSpecies(i);
                }
              }

              //The next value is the spacing number
              sVal = (String) oLine.elementAt(iNumSpecies + 2);
              oPlanting.SetPlantSpacing(new Integer(sVal).intValue());

              //The next value is tree spacing
              sVal = (String) oLine.elementAt(iNumSpecies + 3);
              oPlanting.SetSpacingDistance(new Float(sVal).floatValue());

              //Add grid cells
              p_oCells = Cell.TransformOldGridCell(iGrid, fOldXCellLength,
                  fOldYCellLength,
                  iOldNumXGrids, iOldNumYGrids,
                  oPlot);
              for (i = 0; i < p_oCells.length; i++) {
                oPlanting.AddCell(p_oCells[i]);
              }

              //Add to our list
              oPlantingList = AddPlantingData(oPlantingList, oPlanting);
            }
          }
          oLine = ModelFileFunctions.ReadLine(oIn);
          sVal = (String) oLine.elementAt(0);
          iVal = new Integer(sVal);
        }
      }
      //Now copy the list to our existing list
      mp_oPlantings = oPlantingList;
      Behavior oBeh = GetBehaviorByKey("Planting");
      if (mp_oPlantings.size() > 0) {
        oBeh.m_bIsEnabled = true;
      }
      else {
        oBeh.m_bIsEnabled = false;
      }
    }
    catch (java.io.IOException oErr) {
      ModelException oExp = new ModelException();
      oExp.iErrorCode = ErrorGUI.BAD_FILE;
      oExp.sCallingFunction = "ReadOldHarvestRegimeFile";
      oExp.sMessage = sFileName;
      throw (oExp);
    }
    finally {
      try {
        oIn.close();
      }
      catch (java.io.IOException e) {
        ;
      }
    }
  }

  /**
   * This adds the data contained in a PlantingData object to a list of
   * planting data.  If there is data to be added, it is compared to existing
   * PlantingData objects on the list.  If they match except for the grid cell,
   * then the new object's grid cell is added to the existing object and the
   * new object thrown away.  If it is truly new data, it is added to the list.
   * <p>
   * IMPORTANT:  This assumes that abundances will be in species order, as they
   * should be if reading a harvest regime file.
   * @param oList The list of PlantingData objects to add to.
   * @param oNewPlanting The PlantingData object to add.
   * @return The new list with the data added.
   * @throws ModelException Passing through underlying exceptions.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected Vector AddPlantingData(Vector oList, PlantingData oNewPlanting) throws
      ModelException {
    PlantingData oExistingPlanting;
    int i, j, iNumSpecies;
    boolean bMatch;
    if (null == oNewPlanting) {
      return oList;
    }
    //If the list is currently null, create it
    if (null == oList) {
      oList = new Vector(0);
      oList.add(oNewPlanting);
      return oList;
    }
    for (i = 0; i < oList.size(); i++) {
      oExistingPlanting = (PlantingData) oList.elementAt(i);
      //Compare all the data
      if (oExistingPlanting.GetPlantSpacing() == oNewPlanting.GetPlantSpacing() &&
          oExistingPlanting.GetTimestep() == oNewPlanting.GetTimestep() &&
          oExistingPlanting.GetSpacingDistance() ==
          oNewPlanting.GetSpacingDistance() &&
          oExistingPlanting.GetNumberOfSpecies() ==
          oNewPlanting.GetNumberOfSpecies()) {

        //Do all the species and their abundances match?
        bMatch = true;
        iNumSpecies = oExistingPlanting.GetNumberOfSpecies();
        for (j = 0; j < iNumSpecies; j++) {
          if (oExistingPlanting.GetSpecies(j) != oNewPlanting.GetSpecies(j) ||
              oExistingPlanting.GetAbundance(oExistingPlanting.GetSpecies(j)) !=
              oNewPlanting.GetAbundance(oNewPlanting.GetSpecies(j))) {
            bMatch = false;
            break;
          }
        }

        //They are the same except for cells - add this one's grid cell to the
        //existing list
        if (bMatch) {
          for (j = 0; j < oNewPlanting.GetNumberOfCells(); j++) {
            oExistingPlanting.AddCell(oNewPlanting.GetCell(j));
          }
          return oList;
        }
      }
    }
    //If we've gotten to here, then the new object's data is unique and should
    //be added.
    oList.add(oNewPlanting);
    return oList;
  }

  /**
   * Validates the data prior to writing it.  It causes all PlantingData
   * objects to validate themselves.
   * @param oPop TreePopulation object
   * @throws ModelException if one of the data objects is not valid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void ValidateData(TreePopulation oPop) throws ModelException {
    PlantingData oData;
    Plot oPlot = m_oManager.GetPlot();
    int i, iNumObjects = mp_oPlantings.size();
    for (i = 0; i < iNumObjects; i++) {
      oData = (PlantingData) mp_oPlantings.elementAt(i);
      oData.ValidatePlant(oPop, oPlot);
    }
  }

  /**
   * Overridden because planting parameters are special.
   * @param jOut File to write to.
   * @param oPop TreePopulation object.
   * @throws ModelException if there's a problem writing the file.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void WriteXML(BufferedWriter jOut, TreePopulation oPop) throws
      ModelException {
    try {
      PlantingData oData;
      Cell oCell;
      int i, iNumObjects = mp_oPlantings.size(), j, iTemp;
      if (iNumObjects == 0) {
        return;
      }

      jOut.write("<plant>");

      //Write out the initial diam10s - they may not have been changed but
      //it won't hurt to write them out anyway in that case
      boolean[] p_bUse = new boolean[oPop.GetNumberOfSpecies()];
      for (i = 0; i < p_bUse.length; i++) {p_bUse[i] = true;};
      WriteSpeciesSpecificValue(jOut, mp_fInitialDiam10, oPop, p_bUse);


      //Write each planting object's data
      for (i = 0; i < iNumObjects; i++) {
        oData = (PlantingData) mp_oPlantings.elementAt(i);
        jOut.write("<pl_plantEvent>");
        //Write each species
        iTemp = oData.GetNumberOfSpecies();
        for (j = 0; j < iTemp; j++) {
          jOut.write("<pl_applyToSpecies species=\"" +
                     oPop.GetSpeciesNameFromCode(oData.GetSpecies(j)) + "\"/>");
        }
        //Write the timestep
        jOut.write("<pl_timestep>" + String.valueOf(oData.GetTimestep()) +
                   "</pl_timestep>");

        //Write the plant spacing type
        jOut.write("<pl_spaceType>");
        iTemp = oData.GetPlantSpacing();
        if (iTemp == GRIDDED) {
          jOut.write("gridded");
        }
        else if (iTemp == RANDOM) {
          jOut.write("random");
        }
        else {
          ModelException oErr = new ModelException();
          oErr.iErrorCode = ErrorGUI.BAD_DATA;
          oErr.sCallingFunction = "JAVA";
          oErr.sMessage =
              "When writing parameter file, came across unrecognized " +
              "planting type \"" + String.valueOf(iTemp) + "\".";
          throw oErr;
        }
        jOut.write("</pl_spaceType>");

        //Write the distance or spacing
        jOut.write("<pl_distanceOrDensity>" +
                   String.valueOf(oData.GetSpacingDistance()) +
                   "</pl_distanceOrDensity>");

        //Write the amounts to plant
        jOut.write("<pl_amountToPlant>");

        iTemp = oData.GetNumberOfSpecies();
        for (j = 0; j < iTemp; j++) {

          jOut.write("<pl_atpVal species=\"" +
                     oPop.GetSpeciesNameFromCode(oData.GetSpecies(j)) + "\">" +
                     oData.GetAbundance(oData.GetSpecies(j)) +
                     "</pl_atpVal>");
        }
        jOut.write("</pl_amountToPlant>");

        //Write the list of cells
        iTemp = oData.GetNumberOfCells();
        for (j = 0; j < iTemp; j++) {
          oCell = (Cell) oData.GetCell(j);
          jOut.write("<pl_applyToCell x=\"" + String.valueOf(oCell.GetX())
                     + "\" y=\"" + String.valueOf(oCell.GetY()) + "\"/>");
        }
        jOut.write("</pl_plantEvent>");
      }
      jOut.write("</plant>");
    }
    catch (java.io.IOException e) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_FILE;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = "There was a problem writing the parameter file.";
      throw (oErr);
    }
  }

  /**
   * Sets up the planting results grid.
   * @param oPop Tree population.
   * @throws javawrapper.ModelException If anything goes wrong.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void DoSetup(TreePopulation oPop) throws javawrapper.ModelException {
    int iNumSpecies = oPop.GetNumberOfSpecies(),
        i;

    DataMember[] p_oDataMembers = new DataMember[iNumSpecies];
    //mp_oGrids = new Grid[1];
    Grid oNewGrid;

    String sGridName = "Planting Results";

    //The accessible data members are one for each cut range and species
    for (i = 0; i < iNumSpecies; i++) {

      p_oDataMembers[i] = new DataMember("Planted " +
                                         oPop.GetSpeciesNameFromCode(i).
                                         replace('_', ' '),
                                         "planted" + i,
                                         DataMember.INTEGER);

    }

    //Create our grid
    oNewGrid = new Grid(sGridName, p_oDataMembers, null, 8, 8);

    //Add it to the planting behavior
    oNewGrid = m_oManager.AddGrid(oNewGrid);
    GetBehaviorByKey("Planting").AddGrid(oNewGrid);

    //Default the initial diameter at 10 cm values to the tree population's
    //value
    Float[] p_fDiam10s = new Float[iNumSpecies];
    float fDiam10 = oPop.GetNewSeedlingDiam10();
    for (i = 0; i < iNumSpecies; i++) {
      p_fDiam10s[i] = new Float(fDiam10);
    }
    SetVectorValues(mp_fInitialDiam10, p_fDiam10s);
  }

  /**
   * Ensures data is not displayed in parameter window.  This behavior group
   * has its own window for data entry.
   * @param oPop Not used.
   * @return Null.
   */
  public Vector FormatDataForDisplay(TreePopulation oPop) {
    return null;
  }

  /**
   * Sets a data vector's value.  This function looks for the parent tag
   * "pl_amountToPlant".
   * @param sXMLTag Parent XML tag of data vector whose value is to be set.
   * @param sXMLParentTag The immediate parent tag that sXMLTag is within.
   * @param p_oData Vector of data values appropriate to the data type
   * @param p_sChildXMLTags The XML tags of the child elements
   * @param p_bAppliesTo Array of booleans saying which of the vector values
   * should be set.  This is important in the case of species-specifics - the
   * vector index is the species number but not all species are set.
   * @param oParentAttributes Attributes of parent tag.  May be useful when
   * overridding this for unusual tags.
   * @param p_oAttributes Attributes passed from parser.  This may be needed
   * when overriding this function.  Basic species-specific values are
   * already handled by this function.
   * @return true if the value was set successfully; false if the value could
   * not be found.  (This would not be an error, because I need a way to
   * cycle through the objects until one of the objects comes up with a
   * match.)  If a match to a data object is made via XML tag, but the found
   * object is not a ModelVector, this returns false.
   * @throws ModelException if the value could not be assigned to the data
   * object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>May 24, 2004: Created in beta version (LEM)
   */
  public boolean SetVectorValueByXMLTag(String sXMLTag, String sXMLParentTag,
                                        Vector p_oData,
                                        String[] p_sChildXMLTags,
                                        boolean[] p_bAppliesTo,
                                        Attributes oParentAttributes,
                                        Attributes[] p_oAttributes) throws
      ModelException {
    if (sXMLTag.equals("pl_amountToPlant")) {

      //Get the last planting event
      PlantingData oPlanting = (PlantingData) mp_oPlantings.elementAt(mp_oPlantings.size() - 1);

      int i;
      for (i = 0; i < p_oData.size(); i++) {
        Object oData = p_oData.elementAt(i);
        if (oData != null) {
          float fValue = new Float(String.valueOf(oData)).floatValue();
          oPlanting.AddAbundance(i, fValue);
        }
      }
      return true;
    } else {
      return super.SetVectorValueByXMLTag(sXMLTag, sXMLParentTag, p_oData,
                                        p_sChildXMLTags, p_bAppliesTo,
                                        oParentAttributes, p_oAttributes);

    }
  }


  /**
   * Accepts an XML parent tag (empty, no data) from the parser.  This method
   * watches for the following tags:
   * <ul>
   * <li>pl_applyToCell
   * <li>pl_applyToSpecies
   * <li>pl_plantEvent
   * </ul>
   * @param sXMLTag The XML tag.
   * @param oAttributes The attributes of this object.
   * @throws ModelException if there is a problem reading this data.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>May 12, 2004: Created (LEM)
   */
  public void ReadXMLParentTag(String sXMLTag, Attributes oAttributes) throws
      ModelException {

    if (sXMLTag.equals("pl_applyToCell")) {

      //Get the X and Y values
      int iX = new Integer(oAttributes.getValue("x")).intValue(),
          iY = new Integer(oAttributes.getValue("y")).intValue();

      //Assign them to the newest planting event
      PlantingData oPlanting = (PlantingData) mp_oPlantings.elementAt(
          mp_oPlantings.size() - 1);
      oPlanting.AddCell(iX, iY, m_oManager.GetPlot());

    }
    else if (sXMLTag.equals("pl_applyToSpecies")) {

      //Extract species
      TreePopulation oPop = m_oManager.GetTreePopulation();
      int iSpecies = oPop.GetSpeciesCodeFromName(oAttributes.getValue("species"));

      //Get the last planting event
      PlantingData oPlanting = (PlantingData) mp_oPlantings.elementAt(
          mp_oPlantings.size() - 1);

      //Add this species to that event
      oPlanting.AddSpecies(iSpecies);

    }
    else if (sXMLTag.equals("pl_plantEvent")) {

      //Create a new planting event
      PlantingData oPlanting = new PlantingData();
      mp_oPlantings.add(oPlanting);
    }
    else if (sXMLTag.equals("plant")) {

      //Enable the planting behavior
      GetBehaviorByKey("Planting").m_bIsEnabled = true;
    }
    super.ReadXMLParentTag(sXMLTag, oAttributes);
  }

  /**
   * This method looks for the following tags:
   * <ul>
   * <li>pl_timestep
   * <li>pl_spaceType
   * <li>pl_distanceOrDensity
   * </ul>
   * @param sXMLTag XML tag of data object whose value is to be set.
   * @param sXMLParentTag The immediate parent tag that sXMLTag is within.
   * @param oAttributes Attributes of the object.  Ignored, but may be needed
   * by overriding objects.
   * @param oData Data value appropriate to the data type
   * @return true if the value was set successfully; false if the value could
   * not be found.  (This would not be an error, because I need a way to
   * cycle through the objects until one of the objects comes up with a
   * match.)
   * @throws ModelException if the value could not be assigned to the data
   * object, or if the cut type or cut type amount values are unrecognized.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>May 24 2004: Created (LEM)
   */
  public boolean SetSingleValueByXMLTag(String sXMLTag, String sXMLParentTag,
                                        Attributes oAttributes,
                                        Object oData) throws
      ModelException {
    if (sXMLTag != null) {
      if (sXMLTag.equals("pl_timestep")) {

        //Get the timestep value
        int iTimestep = new Integer(String.valueOf(oData)).intValue();

        //Get the last planting event and set this as the timestep
        PlantingData oPlanting = (PlantingData) mp_oPlantings.elementAt(
            mp_oPlantings.size() - 1);
        oPlanting.SetTimestep(iTimestep);

        return true;
      }
      else if (sXMLTag.equals("pl_spaceType")) {

        //Get the space type string
        String sSpaceType = String.valueOf(oData);

        //Get the last planting event
        PlantingData oPlanting = (PlantingData) mp_oPlantings.elementAt(
            mp_oPlantings.size() - 1);

        if (sSpaceType.equalsIgnoreCase("gridded")) {
          oPlanting.SetPlantSpacing(GRIDDED);
        }
        else if (sSpaceType.equalsIgnoreCase("random")) {
          oPlanting.SetPlantSpacing(RANDOM);
        }
        else {
          //Unrecognized - error
          ModelException oErr = new ModelException();
          oErr.iErrorCode = ErrorGUI.BAD_DATA;
          oErr.sCallingFunction = "JAVA";
          oErr.sMessage = "Planting event in file has an unrecognized plant spacing type - " +
              sSpaceType + ".";
          throw (oErr);
        }

        return true;
      }
      else if (sXMLTag.equals("pl_distanceOrDensity")) {

        //Get the spacing or density value
        float fValue = new Float(String.valueOf(oData)).floatValue();

        //Get the last planting event and set this as the density - it's the
        //same as setting this as the distance if it's gridded
        PlantingData oPlanting = (PlantingData) mp_oPlantings.elementAt(
            mp_oPlantings.size() - 1);
        oPlanting.SetDensity(fValue);

        return true;
      }
    }
    return super.SetSingleValueByXMLTag(sXMLTag, sXMLParentTag, oAttributes, oData);

  }

  /**
   * Checks planting settings upon change of species.  This will
   * remove any deleted species.
   * @param iOldNumSpecies says how many species there used to be.
   * @param p_iIndexer is an array, sized to the new number of species.  For
   * each bucket (representing the index number of a species on the new list),
   * the value is either the index of that same species in the old species
   * list, or -1 if the species is new.
   * @throws ModelException if anything goes wrong.
   */
  public void ChangeOfSpecies(int iOldNumSpecies, int[] p_iIndexer) throws
      ModelException {
    PlantingData oPlanting;
    int[] p_iOldCutSpecies;
    int i, j, k;

    //Go through all the harvests and check for species that have been removed
    for (i = 0; i < mp_oPlantings.size(); i++) {
      oPlanting = (PlantingData) mp_oPlantings.elementAt(i);
      //Make a list of the species for this cut
      p_iOldCutSpecies = new int[oPlanting.GetNumberOfSpecies()];
      for (j = 0; j < p_iOldCutSpecies.length; j++) {
        p_iOldCutSpecies[j] = oPlanting.GetSpecies(j);
      }
      //Now remove all the old species
      oPlanting.ClearSpecies();

      //Now add back the species from the copy we saved
      for (j = 0; j < p_iOldCutSpecies.length; j++) {

        //Look for this species in the new species list - add the new species
        //number if we find it
        for (k = 0; k < p_iIndexer.length; k++) {
          if (p_iIndexer[k] == p_iOldCutSpecies[j]) {
            //Yep, this species is still on the list - change the species
            //number
            oPlanting.AddSpecies(k);
            break;
          }
        }
      }

      //If this harvest event has no species left, delete it
      if (oPlanting.GetNumberOfSpecies() == 0) {
        mp_oPlantings.remove(i);
        i--;
      }
    }
  }
}
