package javawrapper;

/**
 * This class allows you to group together a species number and a type number
 * into one object that can be placed in an array or vector or passed as an
 * argument.  This is useful, for instance, when a behavior object wants to
 * keep a list of those species and types to which each of its sub-behaviors
 * applies.
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author not attributable
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 */

public class SpeciesTypeCombo {
  private int m_iSpecies, /**<Tree species.*/
      m_iType; /**<Tree type.*/

  /**
   * Constructor.
   * @param iSpecies Species number.
   * @param iType Type number.
   * @param oPop Tree population object for data validation.
   * @throws ModelException if either species or type number is invalid
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public SpeciesTypeCombo(int iSpecies, int iType, TreePopulation oPop) throws
      ModelException {

    SetSpecies(iSpecies, oPop);

    if (iType < 0 || iType >= oPop.GetNumberOfTypes()) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_DATA;
      oErr.sMessage = "SpeciesTypeCombo - Invalid type number \"" +
          String.valueOf(iType) + "\".";
      oErr.sCallingFunction = "JAVA";
      throw (oErr);
    }
    m_iType = iType;

  }

  /**
   * Constructor.  Private so that the cloning process can use it without
   * having to go through validation with the TreePopulation - since
   * presumably an already existing object would have validated in its own
   * constructor.
   * @param iSpecies Tree species.
   * @param iType Tree type.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  private SpeciesTypeCombo(int iSpecies, int iType) {
    m_iSpecies = iSpecies;
    m_iType = iType;
  }

  /**
   * Species setter.
   * @param iSpecies Species to set.
   * @param oPop TreePopulation object.
   * @throws ModelException if the species is invalid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetSpecies(int iSpecies, TreePopulation oPop) throws
      ModelException {
    if (iSpecies < 0 || iSpecies >= oPop.GetNumberOfSpecies()) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_DATA;
      oErr.sMessage = "SpeciesTypeCombo - Invalid species number \"" +
          String.valueOf(iSpecies) + "\".";
      oErr.sCallingFunction = "JAVA";
      throw (oErr);
    }
    m_iSpecies = iSpecies;
  }

  /**
   * Implemented equality checker.
   * @param oToCheck Object to check for equality.
   * @return true if species and type both match, false otherwise.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public boolean equals(Object oToCheck) {
    if (! (oToCheck instanceof SpeciesTypeCombo)) {
      return false;
    }

    SpeciesTypeCombo oChecker = (SpeciesTypeCombo) oToCheck;
    if (oChecker.GetSpecies() == m_iSpecies && oChecker.GetType() == m_iType) {
      return true;
    }
    return false;
  }

  /**
   * Override this so clones can be made.
   * @return Clone of this object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public Object clone() {
    SpeciesTypeCombo oClone = new SpeciesTypeCombo(m_iSpecies, m_iType);
    return oClone;
  }

  /**
   * Gets the species for this combo.
   * @return The species.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public int GetSpecies() {
    return m_iSpecies;
  }

  /**
   * Gets the type for this combo.
   * @return The type.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public int GetType() {
    return m_iType;
  }
}