package javawrapper;

/**
 * Manages substrate data and behaviors.
 * <p>Copyright: Copyright (c) 2003 Charles D. Canham</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 * <br>July 21, 2004:  Added calcs grids (LEM)
 * <br>April 1, 2005:  Changed some substrate names per CDC (LEM)
 * <br>April 22, 2005:  Added new snag uproot parameter (LEM)
 * <br>June 3, 2005:  Added combined substrate (LEM)
 * <br>November 17, 2005:  New decay equation for decayed logs (LEM)
 */

public class SubstrateBehaviors
    extends BehaviorTypeBase {

  protected ModelVector
      /**Proportion of dead that fall for each species.*/
      mp_fProportionOfDeadThatFall = new ModelVector(
      "Proportion of Dead that Fall", "su_propOfDeadFall", "su_podfVal", 0,
      ModelVector.FLOAT);

  /**Proportion of fallen that uproot for each species.*/
  protected ModelVector mp_fProportionOfFallenThatUproot = new ModelVector(
      "Proportion of Fallen that Uproot", "su_propOfFallUproot", "su_pofuVal",
      0, ModelVector.FLOAT);

  /**Proportion of snags that uproot for each species. LEM 04/22/05*/
  protected ModelVector mp_fProportionOfSnagsThatUproot = new ModelVector(
      "Proportion of Snags that Uproot", "su_propOfSnagsUproot", "su_posuVal",
      0, ModelVector.FLOAT);

  protected ModelFloat
      /**Scarified soil decay alpha*/
      m_fScarSoilA = new ModelFloat(0, "Scarified Soil Annual Decay Alpha",
                                    "su_scarSoilDecayAlpha");

  /**Scarified soil decay beta*/
  protected ModelFloat m_fScarSoilB = new ModelFloat(0,
      "Scarified Soil Annual Decay Beta",
      "su_scarSoilDecayBeta");

  /**Tip-Up Mounds decay alpha*/
  protected ModelFloat m_fTipUpA = new ModelFloat(0,
                                                  "Tip-Up Mounds Annual Decay Alpha",
                                                  "su_tipupDecayAlpha");

  /**Tip-Up Mounds decay beta*/
  protected ModelFloat m_fTipUpB = new ModelFloat(0,
                                                  "Tip-Up Mounds Annual Decay Beta",
                                                  "su_tipupDecayBeta");

  /**Fresh log decay alpha*/
  protected ModelFloat m_fFreshlogA = new ModelFloat(0,
      "Fresh Log Annual Decay Alpha", "su_freshLogDecayAlpha");

  /**Fresh log decay beta*/
  protected ModelFloat m_fFreshlogB = new ModelFloat(0,
      "Fresh Log Annual Decay Beta", "su_freshLogDecayBeta");

  /**Decayed log decay alpha*/
  protected ModelFloat m_fDecayedlogA = new ModelFloat(0,
      "Decayed Log Annual Decay Alpha", "su_decayedLogDecayAlpha");

  /**Decayed log decay beta*/
  protected ModelFloat m_fDecayedlogB = new ModelFloat(0,
      "Decayed Log Annual Decay Beta", "su_decayedLogDecayBeta");

  /**Scarified soil initial condition proportion*/
  protected ModelFloat m_fInitCondScarSoil = new ModelFloat(0,
      "Initial Conditions Proportion of Scarified Soil", "su_initialScarSoil");

  /**Scarified soil partial cut proportion*/
  protected ModelFloat m_fPartialCutScarSoil = new ModelFloat(0,
      "Partial Cut Proportion of Scarified Soil",
      "su_partialCutScarSoil");

  /**Scarified soil gap cut proportion*/
  protected ModelFloat m_fGapCutScarSoil = new ModelFloat(0,
      "Gap Cut Proportion of Scarified Soil",
      "su_gapCutScarSoil");

  /**Scarified soil clear cut proportion*/
  protected ModelFloat m_fClearCutScarSoil = new ModelFloat(0,
      "Clear Cut Proportion of Scarified Soil",
      "su_clearCutScarSoil");

  /**Tip-Up Mounds initial condition proportion*/
  protected ModelFloat m_fInitCondTipup = new ModelFloat(0,
      "Initial Conditions Proportion of Tip-Up Mounds", "su_initialTipup");

  /**Tip-Up Mounds partial cut proportion*/
  protected ModelFloat m_fPartialCutTipup = new ModelFloat(0,
      "Partial Cut Proportion of Tip-Up Mounds",
      "su_partialCutTipup");

  /**Tip-Up Mounds gap cut proportion*/
  protected ModelFloat m_fGapCutTipup = new ModelFloat(0,
      "Gap Cut Proportion of Tip-Up Mounds",
      "su_gapCutTipup");

  /**Tip-Up Mounds clear cut proportion*/
  protected ModelFloat m_fClearCutTipup = new ModelFloat(0,
      "Clear Cut Proportion of Tip-Up Mounds",
      "su_clearCutTipup");

  /**Fresh log initial condition proportion*/
  protected ModelFloat m_fInitCondFreshLog = new ModelFloat(0,
      "Initial Conditions Proportion of Fresh Logs", "su_initialFreshLog");

  /**Fresh log partial cut proportion*/
  protected ModelFloat m_fPartialCutFreshLog = new ModelFloat(0,
      "Partial Cut Proportion of Fresh Logs", "su_partialCutFreshLog");

  /**Fresh log gap cut proportion*/
  protected ModelFloat m_fGapCutFreshLog = new ModelFloat(0,
      "Gap Cut Proportion of Fresh Logs",
      "su_gapCutFreshLog");

  /**Fresh log clear cut proportion*/
  protected ModelFloat m_fClearCutFreshLog = new ModelFloat(0,
      "Clear Cut Proportion of Fresh Logs",
      "su_clearCutFreshLog");

  /**Decayed log initial condition proportion*/
  protected ModelFloat m_fInitCondDecLog = new ModelFloat(0,
      "Initial Conditions Proportion of Decayed Logs", "su_initialDecayedLog");

  /**Decayed log partial cut proportion*/
  protected ModelFloat m_fPartialCutDecLog = new ModelFloat(0,
      "Partial Cut Proportion of Decayed Logs", "su_partialCutDecayedLog");

  /**Decayed log gap cut proportion*/
  protected ModelFloat m_fGapCutDecLog = new ModelFloat(0,
      "Gap Cut Proportion of Decayed Logs",
      "su_gapCutDecayedLog");

  /**Decayed log clear cut proportion*/
  protected ModelFloat m_fClearCutDecLog = new ModelFloat(0,
      "Clear Cut Proportion of Decayed Logs",
      "su_clearCutDecayedLog");

  /**Root soil disturbance factor*/
  protected ModelFloat m_fRootTipupFactor = new ModelFloat(0,
      "Uprooted Tree Radius Increase Factor for Root Rip-Out",
      "su_rootTipupFactor");

  /**Proportion of litter/moss that is moss*/
  protected ModelFloat m_fMossProportion = new ModelFloat(0,
      "Proportion of Forest Floor Litter/Moss Pool that is Moss",
      "su_mossProportion");

  /**Whether or not tree fall is directional - LEM 06/03/05*/
  protected ModelEnum m_iDirectionalTreeFall =
      new ModelEnum(new int[] {0, 1}
                    ,
                    new String[] {"false", "true"}
                    ,
                    "Use Directional Tree Fall",
                    "su_directionalTreeFall");

  protected ModelInt
      /**Maximum number of years a substrate event hangs around*/
      m_iMaxDecayTime = new ModelInt(0,
                                     "Maximum Number of Years that Decay Occurs",
                                     "su_maxNumberDecayYears");

  /**
   * Constructor
   * @param oManager GUIManager object.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public SubstrateBehaviors(GUIManager oManager) {
    super(oManager, "Substrate", "substrate");

    m_sHelpID = "data.substrate_behaviors";

    //Set up our child behavior vector
    mp_oChildBehaviors = new Behavior[1];
    mp_oChildBehaviors[0] = new Behavior("substrate", "Substrate", "substrate",
                                         (float) 2.1);
    //Substrate cannot apply to seedlings
    mp_oChildBehaviors[0].SetCanApplyTo(TreePopulation.SEEDLING, false);

    mp_oChildBehaviors[0].AddRequiredData(m_iMaxDecayTime);
    mp_oChildBehaviors[0].AddRequiredData(m_fScarSoilA);
    mp_oChildBehaviors[0].AddRequiredData(m_fScarSoilB);
    mp_oChildBehaviors[0].AddRequiredData(m_fTipUpA);
    mp_oChildBehaviors[0].AddRequiredData(m_fTipUpB);
    mp_oChildBehaviors[0].AddRequiredData(m_fFreshlogA);
    mp_oChildBehaviors[0].AddRequiredData(m_fFreshlogB);
    mp_oChildBehaviors[0].AddRequiredData(m_fDecayedlogA);
    mp_oChildBehaviors[0].AddRequiredData(m_fDecayedlogB);
    mp_oChildBehaviors[0].AddRequiredData(m_fMossProportion);
    mp_oChildBehaviors[0].AddRequiredData(m_fRootTipupFactor);
    mp_oChildBehaviors[0].AddRequiredData(m_iDirectionalTreeFall);
    mp_oChildBehaviors[0].AddRequiredData(m_fInitCondScarSoil);
    mp_oChildBehaviors[0].AddRequiredData(m_fInitCondTipup);
    mp_oChildBehaviors[0].AddRequiredData(m_fInitCondFreshLog);
    mp_oChildBehaviors[0].AddRequiredData(m_fInitCondDecLog);
    mp_oChildBehaviors[0].AddRequiredData(m_fPartialCutScarSoil);
    mp_oChildBehaviors[0].AddRequiredData(m_fPartialCutTipup);
    mp_oChildBehaviors[0].AddRequiredData(m_fPartialCutFreshLog);
    mp_oChildBehaviors[0].AddRequiredData(m_fPartialCutDecLog);
    mp_oChildBehaviors[0].AddRequiredData(m_fGapCutScarSoil);
    mp_oChildBehaviors[0].AddRequiredData(m_fGapCutTipup);
    mp_oChildBehaviors[0].AddRequiredData(m_fGapCutFreshLog);
    mp_oChildBehaviors[0].AddRequiredData(m_fGapCutDecLog);
    mp_oChildBehaviors[0].AddRequiredData(m_fClearCutScarSoil);
    mp_oChildBehaviors[0].AddRequiredData(m_fClearCutTipup);
    mp_oChildBehaviors[0].AddRequiredData(m_fClearCutFreshLog);
    mp_oChildBehaviors[0].AddRequiredData(m_fClearCutDecLog);
    mp_oChildBehaviors[0].AddRequiredData(mp_fProportionOfDeadThatFall);
    mp_oChildBehaviors[0].AddRequiredData(mp_fProportionOfFallenThatUproot);
    mp_oChildBehaviors[0].AddRequiredData(mp_fProportionOfSnagsThatUproot);

    mp_oAllData.add(m_fScarSoilA);
    mp_oAllData.add(m_fScarSoilB);
    mp_oAllData.add(m_fTipUpA);
    mp_oAllData.add(m_fTipUpB);
    mp_oAllData.add(m_fFreshlogA);
    mp_oAllData.add(m_fFreshlogB);
    mp_oAllData.add(m_fDecayedlogA);
    mp_oAllData.add(m_fDecayedlogB);
    mp_oAllData.add(m_iMaxDecayTime);
    mp_oAllData.add(m_fInitCondScarSoil);
    mp_oAllData.add(m_fInitCondTipup);
    mp_oAllData.add(m_fInitCondFreshLog);
    mp_oAllData.add(m_fInitCondDecLog);
    mp_oAllData.add(m_fPartialCutScarSoil);
    mp_oAllData.add(m_fPartialCutTipup);
    mp_oAllData.add(m_fPartialCutFreshLog);
    mp_oAllData.add(m_fPartialCutDecLog);
    mp_oAllData.add(m_fGapCutScarSoil);
    mp_oAllData.add(m_fGapCutTipup);
    mp_oAllData.add(m_fGapCutFreshLog);
    mp_oAllData.add(m_fGapCutDecLog);
    mp_oAllData.add(m_fClearCutScarSoil);
    mp_oAllData.add(m_fClearCutTipup);
    mp_oAllData.add(m_fClearCutFreshLog);
    mp_oAllData.add(m_fClearCutDecLog);
    mp_oAllData.add(mp_fProportionOfDeadThatFall);
    mp_oAllData.add(mp_fProportionOfFallenThatUproot);
    mp_oAllData.add(mp_fProportionOfSnagsThatUproot);
    mp_oAllData.add(m_fRootTipupFactor);
    mp_oAllData.add(m_fMossProportion);
    mp_oAllData.add(m_iDirectionalTreeFall);
  }

  /**
   * Sets up the substrate grids.
   * @param oPop Tree population.
   * @throws ModelException
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>July 21, 2004:  Added calcs grids (LEM)
   */
  public void DoSetup(TreePopulation oPop) throws ModelException {
    //mp_oGrids = new Grid[2];
    Grid oNewGrid;

    //************************
     // Regular substrate
     //************************
    String sGridName = "Substrate";

    //Create the six data members
    DataMember[] p_oDataMembers = new DataMember[6];

    p_oDataMembers[0] = new DataMember("Proportion of scarified soil",
                                       "scarsoil",
                                       DataMember.FLOAT);
    p_oDataMembers[1] = new DataMember("Proportion of tip-up mounds",
                                       "tipup", DataMember.FLOAT);
    p_oDataMembers[2] = new DataMember("Proportion of fresh logs", "freshlog",
                                       DataMember.FLOAT);
    p_oDataMembers[3] = new DataMember("Proportion of decayed logs", "declog",
                                       DataMember.FLOAT);
    p_oDataMembers[4] = new DataMember("Proportion of forest floor moss",
                                       "ffmoss",
                                       DataMember.FLOAT);
    p_oDataMembers[5] = new DataMember("Proportion of forest floor litter",
                                       "fflitter",
                                       DataMember.FLOAT);

    DataMember[] p_oPackageDataMembers = new DataMember[4];

    p_oPackageDataMembers[0] = new DataMember("Substrate cohort age", "age",
                                              DataMember.INTEGER);
    p_oPackageDataMembers[1] = new DataMember(
        "Substrate cohort new scarified soil subtrate", "scarsoil",
        DataMember.FLOAT);
    p_oPackageDataMembers[2] = new DataMember(
        "Substrate cohort new tip-up mounds substrate",
        "tipup", DataMember.FLOAT);
    p_oPackageDataMembers[3] = new DataMember("Substrate cohort new fresh logs",
                                              "freshlog",
                                              DataMember.FLOAT);

    oNewGrid = new Grid(sGridName, p_oDataMembers, p_oPackageDataMembers, 8, 8);

    //Add to the substrate behavior
    oNewGrid = m_oManager.AddGrid(oNewGrid);
    GetBehaviorByKey("substrate").AddGrid(oNewGrid);

    //************************
     // Regular substrate calculations grid
     //************************
    sGridName = "substratecalcs";

    //Create the data members
    int iDecayTimesteps = (int) java.lang.Math.ceil( (float) m_iMaxDecayTime.
        GetValue() / m_oManager.GetPlot().GetNumberOfYearsPerTimestep());
    p_oDataMembers = new DataMember[1 + iDecayTimesteps];

    p_oDataMembers[0] = new DataMember("Amount of new tip-up mounds",
                                       "newtipup",
                                       DataMember.FLOAT);
    for (int i = 0; i < iDecayTimesteps; i++) {
      p_oDataMembers[1 +
          i] = new DataMember("Fresh logs " + String.valueOf(i) +
                              " timesteps ago",
                              "freshlog_" + String.valueOf(i),
                              DataMember.FLOAT);
    }
    oNewGrid = new Grid(sGridName, p_oDataMembers, null, 8, 8);
    //Add to the substrate behavior
    oNewGrid = m_oManager.AddGrid(oNewGrid);
    GetBehaviorByKey("substrate").AddGrid(oNewGrid);

    //************************
     // Defaults
     //************************
      //Set the proportion of snags that uproot to all 0s, in case snags aren't
      //being used
    int iNumSpecies = oPop.GetNumberOfSpecies(), i;
    if (mp_fProportionOfSnagsThatUproot.GetValue().size() == 0) {
      for (i = 0; i < iNumSpecies; i++) {
        mp_fProportionOfSnagsThatUproot.GetValue().add(new Float(0));
      }
    }
  }

  /**
   * Overridden from WorkerBase.  This is on the lookout for changes to
   * m_iMaxDecayTime, so DoSetup() can be triggered and the grids recreated.
   * @param oDataMember Data member to set.
   * @param oData Data value to set into data member.
   * @throws ModelException if data is of the wrong type.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected void SetSingleValue(ModelData oDataMember, Object oData) throws
      ModelException {
    super.SetSingleValue(oDataMember, oData);
    if (oDataMember.equals(m_iMaxDecayTime)) {
      DoSetup(m_oManager.GetTreePopulation());
    }
  }

  /**
   * Overridden to provide backwards compatibility - I changed XML tag
   * names.  This was put in for 6.05 - remove after several more versions.
   * @param sXMLTag XML tag of data object whose value is to be set.
   * @param sXMLParentTag The immediate parent tag that sXMLTag is within.
   * @param oAttributes Attributes of the object.  Ignored, but may be needed
   * by overriding objects.
   * @param oData Data value, either String or type appropriate to the data
   * type
   * @return true if the value was set successfully; false if the value could
   * not be found.
   * @throws ModelException if the value could not be assigned to the data
   * object.
   */
  public boolean SetSingleValueByXMLTag(String sXMLTag, String sXMLParentTag,
                                        org.xml.sax.Attributes oAttributes,
                                        Object oData) throws
      ModelException {

    if (sXMLParentTag.equals(m_sXMLTag)) {
      if (sXMLTag.equals("su_DecayedLogToForestFloor") ||
          sXMLTag.equals("su_mossyLogToCanMoss")) {
        //Old straight proportion of decayed log decay - transform it to
        //a and b
        Float oFloat;
        if (oData instanceof String) {
          oFloat = new Float( (String) oData);
        }
        else {
          oFloat = (Float) oData;
        }
        m_fDecayedlogA.SetValue((float)java.lang.Math.log((1 - oFloat.floatValue())));
        m_fDecayedlogB.SetValue(1);
        return true;
      }
      else if (sXMLTag.equals("su_gapmossDecayAlpha")) {
        sXMLTag = "su_scarSoilDecayAlpha";
      }
      else if (sXMLTag.equals("su_gapmossDecayBeta")) {
        sXMLTag = "su_scarSoilDecayBeta";
      }
      else if (sXMLTag.equals("su_organicDecayAlpha")) {
        sXMLTag = "su_tipupDecayAlpha";
      }
      else if (sXMLTag.equals("su_organicDecayBeta")) {
        sXMLTag = "su_tipupDecayBeta";
      }
      else if (sXMLTag.equals("su_initialGapMoss")) {
        sXMLTag = "su_initialScarSoil";
      }
      else if (sXMLTag.equals("su_initialOrganic")) {
        sXMLTag = "su_initialTipup";
      }
      else if (sXMLTag.equals("su_initialMossyLog")) {
        sXMLTag = "su_initialDecayedLog";
      }
      else if (sXMLTag.equals("su_partialCutGapMoss")) {
        sXMLTag = "su_partialCutScarSoil";
      }
      else if (sXMLTag.equals("su_partialCutOrganic")) {
        sXMLTag = "su_partialCutTipup";
      }
      else if (sXMLTag.equals("su_partialCutMossyLog")) {
        sXMLTag = "su_partialCutDecayedLog";
      }
      else if (sXMLTag.equals("su_gapCutGapMoss")) {
        sXMLTag = "su_gapCutScarSoil";
      }
      else if (sXMLTag.equals("su_gapCutOrganic")) {
        sXMLTag = "su_gapCutTipup";
      }
      else if (sXMLTag.equals("su_gapCutMossyLog")) {
        sXMLTag = "su_gapCutDecayedLog";
      }
      else if (sXMLTag.equals("su_clearCutGapMoss")) {
        sXMLTag = "su_clearCutScarSoil";
      }
      else if (sXMLTag.equals("su_clearCutOrganic")) {
        sXMLTag = "su_clearCutTipup";
      }
      else if (sXMLTag.equals("su_clearCutMossyLog")) {
        sXMLTag = "su_clearCutDecayedLog";
      }
    }
    //Grid values
    /*else if (sXMLTag.equals("ma_floatCode")) {
      if (m_iGridParseIndex > -1) {
        String sLabel = oAttributes.getValue("label");
        boolean bUpdate = false;
        if (sLabel.equals("gapmoss")) {
          sLabel = "scarsoil";
          bUpdate = true;
        }
        else if (sLabel.equals("canmoss")) {
          sLabel = "ffmoss";
          bUpdate = true;
        }
        else if (sLabel.equals("organic")) {
          sLabel = "tipup";
          bUpdate = true;
        }
        else if (sLabel.equals("mossylog")) {
          sLabel = "declog";
          bUpdate = true;
        }
        //We can't set the value in oAttributes and call super, because
        //oAttributes won't let us; so we have to copy the setting code
        //here
        if (true == bUpdate) {
          int iIndex = new Integer( (String) oData).intValue();
          if (m_iPackageIndex > -1) {
            mp_oGrids[m_iGridParseIndex].SetGridPackageFloatCode(sLabel,
                iIndex);
            return true;
          }
          else {
            mp_oGrids[m_iGridParseIndex].SetGridFloatCode(sLabel, iIndex);
            return true;
          }
        }
      }
    } */
    return super.SetSingleValueByXMLTag(sXMLTag, sXMLParentTag, oAttributes,
                                        oData);
  }

  /**
   * Validates the data in preparation for parameter file writing or some such.
   * @throws ModelException if:
   * <ul>
   * <li>m_fRootTipupFactor or m_iMaxDecayTime are not positive numbers</li>
   * <li>if the values in mp_fProportionOfDeadThatFall,
   * mp_fProportionOfFallenThatUproot, and mp_fProportionOfSnagsThatUproot are
   * not proportions</li>
   * <li>m_fMossProportion isn't a proportion</li>
   * <li>any initial or harvest conditions value isn't a proportion</li>
   * <li>the total of any set of initial or harvest conditions is greater than
   * 1</li>
   * </ul>
   * @param oPop Not used.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   * <br>April 22, 2005:  Added proportion checking (LEM)
   */
  public void ValidateData(TreePopulation oPop) throws ModelException {

    if (GetBehaviorByKey("substrate").m_bIsEnabled) {
      boolean[] p_bApplies = GetBehaviorByKey("substrate").GetWhichSpeciesUsed(
          oPop);

      MakeSurePositive(m_fRootTipupFactor);
      MakeSurePositive(m_iMaxDecayTime);
      MakeSureIsProportion(m_fMossProportion);
      MakeSureAllAreProportions(mp_fProportionOfDeadThatFall, p_bApplies);
      MakeSureAllAreProportions(mp_fProportionOfFallenThatUproot, p_bApplies);
      MakeSureAllAreProportions(mp_fProportionOfSnagsThatUproot, p_bApplies);
      MakeSureIsProportion(m_fInitCondDecLog);
      MakeSureIsProportion(m_fInitCondFreshLog);
      MakeSureIsProportion(m_fInitCondScarSoil);
      MakeSureIsProportion(m_fInitCondTipup);
      MakeSureIsProportion(m_fPartialCutDecLog);
      MakeSureIsProportion(m_fPartialCutFreshLog);
      MakeSureIsProportion(m_fPartialCutScarSoil);
      MakeSureIsProportion(m_fPartialCutTipup);
      MakeSureIsProportion(m_fGapCutDecLog);
      MakeSureIsProportion(m_fGapCutFreshLog);
      MakeSureIsProportion(m_fGapCutScarSoil);
      MakeSureIsProportion(m_fGapCutTipup);
      MakeSureIsProportion(m_fClearCutDecLog);
      MakeSureIsProportion(m_fClearCutFreshLog);
      MakeSureIsProportion(m_fClearCutScarSoil);
      MakeSureIsProportion(m_fClearCutTipup);

      if (m_fInitCondDecLog.GetValue() +
          m_fInitCondFreshLog.GetValue() +
          m_fInitCondScarSoil.GetValue() +
          m_fInitCondTipup.GetValue() > 1) {
        throw new ModelException(ErrorGUI.BAD_DATA, "JAVA",
            "The initial conditions values must not add up to be greater than 1.");
      }

      if (m_fPartialCutDecLog.GetValue() +
          m_fPartialCutFreshLog.GetValue() +
          m_fPartialCutScarSoil.GetValue() +
          m_fPartialCutTipup.GetValue() > 1) {
        throw new ModelException(ErrorGUI.BAD_DATA, "JAVA",
            "The partial cut values must not add up to be greater than 1.");
      }

      if (m_fGapCutDecLog.GetValue() +
          m_fGapCutFreshLog.GetValue() +
          m_fGapCutScarSoil.GetValue() +
          m_fGapCutTipup.GetValue() > 1) {
        throw new ModelException(ErrorGUI.BAD_DATA, "JAVA",
            "The gap cut values must not add up to be greater than 1.");
      }

      if (m_fClearCutDecLog.GetValue() +
          m_fClearCutFreshLog.GetValue() +
          m_fClearCutScarSoil.GetValue() +
          m_fClearCutTipup.GetValue() > 1) {
        throw new ModelException(ErrorGUI.BAD_DATA, "JAVA",
            "The clear cut values must not add up to be greater than 1.");
      }
    }
  }
}
