package javawrapper;

import java.io.FileInputStream;
import com.ice.tar.TarInputStream;
import com.ice.tar.TarEntry;
import com.ice.tar.TarArchive;
import com.ice.tar.TarHeader;
import java.io.File;
import java.util.zip.GZIPInputStream;
import java.io.InputStream;
import java.io.FileOutputStream;

/**
 * Functions for working with tarballs (.gz.tar files).
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 */

public class Tarball {
  /**
   * Extracts a tarball's entries into the designated temp folder.  If there is
   * file info in the tar archive, they automatically get put in a sub folder
   * of the root directory.
   * @param sTarball Tarball to extract
   * @param sTempRoot Temp root directory
   * @return File where the files were extracted to.
   * @throws ModelException wraps IOExceptions.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public static String ExtractTarball(String sTarball, String sTempRoot) throws
      ModelException {
    FileInputStream oFileInputStream = null;
    TarInputStream oTarballStream = null;
    try {

      //Find out what directory the output files will be written to
      TarEntry oTarEntry;
      oFileInputStream = new FileInputStream(sTarball);
      oTarballStream = new TarInputStream(oFileInputStream);
      oTarEntry = oTarballStream.getNextEntry();
      String sSeparator = System.getProperty("file.separator"), sDir;

      if (oTarEntry != null) {

        TarHeader oHeader = oTarEntry.getHeader();
        String sFilePath = oHeader.getName();

        if (sFilePath.lastIndexOf(sSeparator) >= 0) {
          sDir = sFilePath.substring(0,
                                     sFilePath.lastIndexOf(sSeparator));
          sDir = sTempRoot + sDir + sSeparator;
        }
        else {
          sDir = sTempRoot + sSeparator;
        }

      }
      else {

        //This file is empty - throw an error
        ModelException oErr = new ModelException();
        oErr.iErrorCode = ErrorGUI.BAD_FILE;
        oErr.sCallingFunction = "JAVA";
        oErr.sMessage = "The file " + sTarball + " contains no data.";
        throw (oErr);

      }

      //Extract the contents to the root - they'll end up in m_sTempDir
      oFileInputStream.close();
      oFileInputStream = new FileInputStream(sTarball);
      TarArchive oTarball = new TarArchive(oFileInputStream);
      oTarball.setKeepOldFiles(false); //overwrite anything already there

      oTarball.extractContents(new File(sTempRoot + sSeparator));

      oTarball.closeArchive();

      return sDir;
    }
    catch (java.io.IOException e) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_FILE;
      oErr.sCallingFunction = "ExtractTarball";
      oErr.sMessage = sTarball;
      throw (oErr);
    }
    finally {
      try {
        if (oFileInputStream != null) {
          oFileInputStream.close();
        }
        if (oTarballStream != null) {
          oTarballStream.close();
        }
      }
      catch (java.io.IOException e) {
        ; // do nothing
      }
    }
  }

  /**
   * This converts a gzipped file to an uncompressed file stream.
   * @param sFileToGet - the file to uncompress.
   * @return - InputStream - the uncompressed file stream
   * @throws ModelException if the file is not found, or wrapping another
   * exception.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public static InputStream UnzipFile(String sFileToGet) throws
      ModelException {
    try {

      GZIPInputStream oGZipStream = new GZIPInputStream(new FileInputStream(new
          File(sFileToGet)));

      return oGZipStream;

    }
    catch (java.io.IOException e) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_FILE;
      oErr.sCallingFunction = "JAVA";
      oErr.sMessage = e.getMessage();
      throw (oErr);
    }
  }

  /**
   * Extracts a file from .gz.tar land to plaintext.  The file will be placed
   * in the same directory as the tarball.
   * @param sTarball Tarball in which the file is.
   * @param sFileName Name of file to extract, with no path information and
   * with no .gz or .tar extension (but with its plaintext extension).
   * @return The full filename and path of the extracted file, or null if the
   * file was not found in the tarball.
   * @throws ModelException if the file is bad.
   */
  public static String ExtractTarballFile(String sTarball, String sFileName) throws
      ModelException {
    TarInputStream oTarballStream = null;
    FileInputStream oInStream = null;
    FileInputStream oZipFileStream = null;
    GZIPInputStream oGZipStream = null;
    try {

      if (sFileName == null || sFileName.length() == 0) {
        return null;
      }

      String sExtractedFile = null;
      TarEntry oTarEntry;
      TarHeader oHeader;
      oInStream = new FileInputStream(sTarball);
      oTarballStream = new TarInputStream(oInStream);
      oTarEntry = oTarballStream.getNextEntry();
      String sSeparator = System.getProperty("file.separator"), sTarFile = "",
          sFilePath = "";

      //Find the tar entry corresponding to this file
      while (oTarEntry != null) {

        oHeader = oTarEntry.getHeader();
        sTarFile = oHeader.getName();

        //Separate out the filename from any path information
        sTarFile = sTarFile.substring(sTarFile.lastIndexOf(sSeparator) + 1);
        if (sTarFile.startsWith(sFileName)) {

          //Get the filepath of the tarball
          sFilePath = sTarball.substring(0, sTarball.lastIndexOf(sSeparator));
          break;
        }

        oTarEntry = oTarballStream.getNextEntry();
      }

      if (oTarEntry == null) {
        return null;
      }

      //Get the new file to which to write the filestream
      sExtractedFile = sFilePath + sSeparator + sTarFile;

      //Extract the file, still gzipped
      ExtractFile(sTarball, oTarEntry, new File(sExtractedFile));

      //Now unzip the file
      //Open the GZIP stream with the file still having the .gz extension
      File oZipFile = new File(sExtractedFile);
      oZipFileStream = new FileInputStream(oZipFile);
      oGZipStream = new GZIPInputStream(oZipFileStream);
      sExtractedFile = sFilePath + sSeparator + sFileName;
      java.io.FileOutputStream oUnzippedOutFile = new java.io.FileOutputStream(
          sExtractedFile);
      int c;
      while ( (c = oGZipStream.read()) != -1) {
        oUnzippedOutFile.write(c);
      }
      oUnzippedOutFile.close();
      oUnzippedOutFile = null;
      oGZipStream.close();
      oGZipStream = null;
      oZipFileStream.close();
      oZipFileStream = null;

      //Delete the gzip file, if it's still there
      File oTemp = new File(sExtractedFile + ".gz");
      oTemp.delete();

      return sExtractedFile;

    }
    catch (java.io.IOException e) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_FILE;
      oErr.sCallingFunction = "ExtractTarballFile";
      oErr.sMessage = e.getMessage();
      throw (oErr);
    }
    finally {
      if (oTarballStream != null) {
        try {
          oTarballStream.close();
        }
        catch (java.io.IOException e) {
          ;
        }
      }
      if (oInStream != null) {
        try {
          oInStream.close();
        }
        catch (java.io.IOException e) {
          ;
        }
      }
      if (oZipFileStream != null) {
        try {
          oZipFileStream.close();
        }
        catch (java.io.IOException e) {
          ;
        }
      }
      if (oGZipStream != null) {
        try {
          oGZipStream.close();
        }
        catch (java.io.IOException e) {
          ;
        }
      }

    }
  }

  /**
   * Deletes all files from the tarball in the temp folder.
   * @param sTarball Tarball with the files to delete.
   * @param sTempDir Directory from which files should be cleaned
   * @param sTempRoot Temp directory root
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public static void CleanUp(String sTarball, String sTempDir, String sTempRoot) {
    FileInputStream oFileInputStream = null;
    TarInputStream oTarballStream = null;
    try {
      //Look through the tarball and delete each file represented by an entry
      File oFile;
      TarEntry oTarEntry;
      oFileInputStream = new FileInputStream(sTarball);
      oTarballStream = new TarInputStream(oFileInputStream);
      oTarEntry = oTarballStream.getNextEntry();
      String sSeparator = System.getProperty("file.separator");

      while (oTarEntry != null) {
        if (oTarEntry.getName().startsWith(sSeparator)) {
          oFile = new File(sTempRoot + oTarEntry.getName());
        }
        else {
          oFile = new File(sTempRoot + sSeparator + oTarEntry.getName());
        }
        oFile.delete();

        oTarEntry = oTarballStream.getNextEntry();
      }

      //If the temp directory is different from the file root, delete it too
      if (!sTempDir.equals(sTempRoot + sSeparator)) {
        while (!sTempDir.equals(sTempRoot)) {
          oFile = new File(sTempDir);
          oFile.delete();
          sTempDir = oFile.getParent();
        }
      }
    }
    catch (java.io.IOException e) {
      //Do nothing
    }
    catch (java.lang.Throwable e) {
      //Do nothing
    }
    finally {
      try {
        if (oFileInputStream != null) {
          oFileInputStream.close();
        }
        if (oTarballStream != null) {
          oTarballStream.close();
        }
      }
      catch (java.io.IOException e) {
        ; //do nothing
      }
    }
  }

  /**
   * Gets a list of file entries in the tarball.  The list is of the core
   * XML files, with no paths attached.
   * @param sTarball Tarball for which to get entries.
   * @return An array of filenames in the tarball, no path info, no .gz
   * extension, or null if the tarball is empty.
   * @throws ModelException if there is a problem reading the file.
   */
  public static String[] GetTarballEntries(String sTarball) throws
      ModelException {

    try {

      TarEntry oTarEntry;
      TarHeader oHeader;
      //Declare a vector to hold entries from the tar stream
      java.util.Vector p_sTempNames = new java.util.Vector(0);
      TarInputStream oTarballStream = new TarInputStream(new FileInputStream(
          sTarball));
      oTarEntry = oTarballStream.getNextEntry();
      String sSeparator = System.getProperty("file.separator"), sFileName;

      //Loop through all the tar entries and extract the filenames
      while (oTarEntry != null) {

        oHeader = oTarEntry.getHeader();
        sFileName = oHeader.getName();

        //Separate out the filename from any path information
        sFileName = sFileName.substring(sFileName.lastIndexOf(sSeparator) + 1);

        //Remove the ".gz" if it's there
        sFileName = sFileName.substring(0, sFileName.indexOf(".gz"));

        p_sTempNames.add(sFileName);

        oTarEntry = oTarballStream.getNextEntry();
      }

      //Now translate the vector into the array to return
      if (p_sTempNames.size() == 0) {
        return null;
      }
      String[] p_sNames = new String[p_sTempNames.size()];
      for (int i = 0; i < p_sNames.length; i++) {
        p_sNames[i] = (String) p_sTempNames.elementAt(i);
      }

      return p_sNames;
    }
    catch (java.io.IOException e) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_FILE;
      oErr.sCallingFunction = "ExtractTarball";
      oErr.sMessage = sTarball;
      throw (oErr);
    }
  }

  /**
   * Extracts a single file from a tar archive.  I grabbed and modified this
   * from the TarTool package - method
   * ice.com.tartool.ArchiveTreePanel.extractFile. That code is in the public
   * domain.  The original author is Timothy Gerard Endres.
   * <p>
   * I stripped out the code that didn't have anything to do with how our files
   * work.  The original code supported MIME types and ascii files, and we will
   * never be bothering with either.
   * <p>
   * When I say that this method "extracts" a file, what it does is copies the
   * contents of the file out of the tar archive and into a destination file.
   * The original tar archive is unchanged and still contains the extracted
   * file.
   * @param sTarball The filename of the tar file.
   * @param oEntry The TarEntry corresponding to the file to extract.
   * @param oDestFile The file to which to write the extracted file's contents.
   * @throws java.io.IOException if there is a problem writing the file.
   */
  private static void
      ExtractFile(
      String sTarball, TarEntry oEntry, File oDestFile) throws java.io.
      IOException {
    FileOutputStream out = null;
    TarInputStream tarIn = null;
    FileInputStream oInputStream = null;

    try {
      oInputStream = new FileInputStream(new File(sTarball));
      tarIn = new TarInputStream(oInputStream);

      for (; ; ) {
        TarEntry e = tarIn.getNextEntry();

        if (e == null) {
          break;
        }

        boolean doIt = false;

        doIt = e.equals(oEntry);
        if (doIt) {
          out = new FileOutputStream(oDestFile);
        }

        if (doIt) {
          byte[] buf = new byte[32 * 1024];

          for (; ; ) {
            int numRead = tarIn.read(buf);

            if (numRead == -1) {
              break;
            }

            out.write(buf, 0, numRead);

          }

          if (out != null) {
            out.close();
            out = null;
          }

          break;
        }
      }
    }
    finally {
      if (out != null) {
        out.close();

      }
      if (tarIn != null) {
        tarIn.close();
      }
      if (oInputStream != null) {
        oInputStream.close();
      }
    }
  }

}
