package javawrapper;

import junit.framework.*;

/**
 * Tests the Harvest and Natural Disturbance behaviors of the
 * DisturbanceBehaviors class.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)

 */
public class TestSelectionHarvest
    extends TestCase {

  /**
   * Tests validation of selection harvest data.
   * @throws ModelException if a test is failed.
   */
  public void testValidateData() throws ModelException {
    GUIManager oManager = null;
    String sFileName = null;
    DisturbanceBehaviors oDist = null;
    try {
      oManager = new GUIManager(null);

      oManager.ClearCurrentData();
      sFileName = WriteXMLFile1();
      oManager.InputXMLParameterFile(sFileName);
      oDist = oManager.GetDisturbanceBehaviors();

      //Make sure that it passes validation
      oDist.ValidateData(oManager.GetTreePopulation());
      assertTrue(oDist.GetBehaviorByKey("Harvest").m_bIsEnabled);
    }
    catch (ModelException oErr) {
      fail("Selection harvest validation failed with message " + oErr.sMessage);
    }
    catch (java.io.IOException oE) {
      fail("Caught IOException.  Message: " + oE.getMessage());
    }
    finally {
      new java.io.File(sFileName).delete();
    }

    //Negative basal area for cut range 1
    oDist.m_fCut1BA.SetValue(-1);
    try {
      oDist.ValidateData(oManager.GetTreePopulation());
      fail("Selection harvest failed to catch negative BA.");
    } catch (ModelException oErr) {;}
    oDist.m_fCut1BA.SetValue(10);
    oDist.ValidateData(oManager.GetTreePopulation());

    //Negative basal area for cut range 2
    oDist.m_fCut2BA.SetValue(-1);
    try {
      oDist.ValidateData(oManager.GetTreePopulation());
      fail("Selection harvest failed to catch negative BA.");
    } catch (ModelException oErr) {;}
    oDist.m_fCut2BA.SetValue(10);
    oDist.ValidateData(oManager.GetTreePopulation());

    //Negative basal area for cut range 3
    oDist.m_fCut3BA.SetValue(-1);
    try {
      oDist.ValidateData(oManager.GetTreePopulation());
      fail("Selection harvest failed to catch negative BA.");
    } catch (ModelException oErr) {;}
    oDist.m_fCut3BA.SetValue(10);
    oDist.ValidateData(oManager.GetTreePopulation());

    //Negative basal area for cut range 4
    oDist.m_fCut4BA.SetValue(-1);
    try {
      oDist.ValidateData(oManager.GetTreePopulation());
      fail("Selection harvest failed to catch negative BA.");
    } catch (ModelException oErr) {;}
    oDist.m_fCut4BA.SetValue(10);
    oDist.ValidateData(oManager.GetTreePopulation());

    //Overlapping range for cut range 1
    oDist.m_fCut1LoDBH.SetValue(100);
    try {
      oDist.ValidateData(oManager.GetTreePopulation());
      fail("Selection harvest failed to catch bad cut range.");
    } catch (ModelException oErr) {;}
    oDist.m_fCut1LoDBH.SetValue(0);
    oDist.ValidateData(oManager.GetTreePopulation());

    //Overlapping range for cut range 2
    oDist.m_fCut2LoDBH.SetValue(100);
    try {
      oDist.ValidateData(oManager.GetTreePopulation());
      fail("Selection harvest failed to catch bad cut range.");
    } catch (ModelException oErr) {;}
    oDist.m_fCut2LoDBH.SetValue(0);
    oDist.ValidateData(oManager.GetTreePopulation());

    //Overlapping range for cut range 3
    oDist.m_fCut3LoDBH.SetValue(100);
    try {
      oDist.ValidateData(oManager.GetTreePopulation());
      fail("Selection harvest failed to catch bad cut range.");
    } catch (ModelException oErr) {;}
    oDist.m_fCut3LoDBH.SetValue(0);
    oDist.ValidateData(oManager.GetTreePopulation());

    //Overlapping range for cut range 4
    oDist.m_fCut4LoDBH.SetValue(100);
    try {
      oDist.ValidateData(oManager.GetTreePopulation());
      fail("Selection harvest failed to catch bad cut range.");
    } catch (ModelException oErr) {;}
    oDist.m_fCut4LoDBH.SetValue(0);
    oDist.ValidateData(oManager.GetTreePopulation());


  }

  /**
   * Tests parameter file reading.
   */
  public void testReadXML() {
    GUIManager oManager = null;
    String sFileName = null;
    try {
      oManager = new GUIManager(null);

      oManager.ClearCurrentData();
      sFileName = WriteXMLFile1();
      oManager.InputXMLParameterFile(sFileName);
      DisturbanceBehaviors oDist = oManager.GetDisturbanceBehaviors();
      assertEquals(5000, oDist.m_iInitialAge.GetValue());

      assertEquals(500, oDist.m_fCut1BA.GetValue(), 0.001);
      assertEquals(0, oDist.m_fCut1LoDBH.GetValue(), 0.001);
      assertEquals(10, oDist.m_fCut1HiDBH.GetValue(), 0.001);

      assertEquals(5, oDist.m_fCut2BA.GetValue(), 0.001);
      assertEquals(10, oDist.m_fCut2LoDBH.GetValue(), 0.001);
      assertEquals(20, oDist.m_fCut2HiDBH.GetValue(), 0.001);

      assertEquals(50, oDist.m_fCut3BA.GetValue(), 0.001);
      assertEquals(21, oDist.m_fCut3LoDBH.GetValue(), 0.001);
      assertEquals(30, oDist.m_fCut3HiDBH.GetValue(), 0.001);

      assertEquals(5000, oDist.m_fCut4BA.GetValue(), 0.001);
      assertEquals(31, oDist.m_fCut4LoDBH.GetValue(), 0.001);
      assertEquals(40, oDist.m_fCut4HiDBH.GetValue(), 0.001);

      new java.io.File(sFileName).delete();
    }
    catch (ModelException oErr) {
      fail("Selection harvest parameter file reading failed with message " + oErr.sMessage);
    }
    catch (java.io.IOException oE) {
      fail("Caught IOException.  Message: " + oE.getMessage());
    }

  }

  /**
   * This writes an XML file to test value setting.
   * @return Name of file.
   * @throws java.io.IOException if there is an IO problem.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  protected String WriteXMLFile1() throws java.io.IOException {
      String sFileName = "\\loratestxml1.xml";
      java.io.FileWriter oOut = new java.io.FileWriter(sFileName);

      oOut.write(
          "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?>");
      oOut.write("<paramFile fileCode=\"06010101\">");

      //Plot
      oOut.write("<plot>");
      oOut.write("<plot_lenX>200.0</plot_lenX>");
      oOut.write("<plot_lenY>200.0</plot_lenY>");
      oOut.write("<plot_latitude>55.37</plot_latitude>");
      oOut.write("</plot>");
      oOut.write("<trees>");
      oOut.write("<tr_speciesList>");
      oOut.write("<tr_species speciesName=\"Species_1\" />");
      oOut.write("<tr_species speciesName=\"Species_2\" />");
      oOut.write("</tr_speciesList>");
      oOut.write("<tr_seedDiam10Cm>0.1</tr_seedDiam10Cm>");
      oOut.write("<tr_canopyHeight>");
      oOut.write("<tr_chVal species=\"Species_1\">39.48</tr_chVal>");
      oOut.write("<tr_chVal species=\"Species_2\">39.48</tr_chVal>");
      oOut.write("</tr_canopyHeight>");
      oOut.write("<tr_stdAsympCrownRad>");
      oOut.write("<tr_sacrVal species=\"Species_1\">0.0549</tr_sacrVal>");
      oOut.write("<tr_sacrVal species=\"Species_2\">0.0549</tr_sacrVal>");
      oOut.write("</tr_stdAsympCrownRad>");
      oOut.write("<tr_stdCrownRadExp>");
      oOut.write("<tr_screVal species=\"Species_1\">1.0</tr_screVal>");
      oOut.write("<tr_screVal species=\"Species_2\">1.0</tr_screVal>");
      oOut.write("</tr_stdCrownRadExp>");
      oOut.write("<tr_conversionDiam10ToDBH>");
      oOut.write("<tr_cdtdVal species=\"Species_1\">0.8008</tr_cdtdVal>");
      oOut.write("<tr_cdtdVal species=\"Species_2\">0.8008</tr_cdtdVal>");
      oOut.write("</tr_conversionDiam10ToDBH>");
      oOut.write("<tr_minAdultDBH>");
      oOut.write("<tr_madVal species=\"Species_1\">10.0</tr_madVal>");
      oOut.write("<tr_madVal species=\"Species_2\">10.0</tr_madVal>");
      oOut.write("</tr_minAdultDBH>");
      oOut.write("</trees>");
      oOut.write("<allometry>");
      oOut.write("<tr_stdAsympCrownHt>");
      oOut.write("<tr_sachVal species=\"Species_1\">0.389</tr_sachVal>");
      oOut.write("<tr_sachVal species=\"Species_2\">0.389</tr_sachVal>");
      oOut.write("</tr_stdAsympCrownHt>");
      oOut.write("<tr_stdCrownHtExp>");
      oOut.write("<tr_scheVal species=\"Species_1\">1.0</tr_scheVal>");
      oOut.write("<tr_scheVal species=\"Species_2\">1.0</tr_scheVal>");
      oOut.write("</tr_stdCrownHtExp>");
      oOut.write("<tr_slopeOfHeight-Diam10>");
      oOut.write("<tr_sohdVal species=\"Species_1\">0.03418</tr_sohdVal>");
      oOut.write("<tr_sohdVal species=\"Species_2\">0.03418</tr_sohdVal>");
      oOut.write("</tr_slopeOfHeight-Diam10>");
      oOut.write("<tr_slopeOfAsymHeight>");
      oOut.write("<tr_soahVal species=\"Species_1\">0.0299</tr_soahVal>");
      oOut.write("<tr_soahVal species=\"Species_2\">0.0299</tr_soahVal>");
      oOut.write("</tr_slopeOfAsymHeight>");
      oOut.write("</allometry>");
      oOut.write("<behaviorList>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>SelectionHarvest</behaviorName>");
      oOut.write("<version>1</version>");
      oOut.write("</behavior>");
      oOut.write("</behaviorList>");
      oOut.write("<SelectionHarvest>");
      oOut.write("<sha_CutRange>");
      oOut.write("<sha_target_BA>500</sha_target_BA>");
      oOut.write("<sha_loDBH>0</sha_loDBH>");
      oOut.write("<sha_hiDBH>10</sha_hiDBH>");
      oOut.write("</sha_CutRange>");
      oOut.write("<sha_CutRange>");
      oOut.write("<sha_target_BA>5</sha_target_BA>");
      oOut.write("<sha_loDBH>10</sha_loDBH>");
      oOut.write("<sha_hiDBH>20</sha_hiDBH>");
      oOut.write("</sha_CutRange>");
      oOut.write("<sha_CutRange>");
      oOut.write("<sha_target_BA>50</sha_target_BA>");
      oOut.write("<sha_loDBH>21</sha_loDBH>");
      oOut.write("<sha_hiDBH>30</sha_hiDBH>");
      oOut.write("</sha_CutRange>");
      oOut.write("<sha_CutRange>");
      oOut.write("<sha_target_BA>5000</sha_target_BA>");
      oOut.write("<sha_loDBH>31</sha_loDBH>");
      oOut.write("<sha_hiDBH>40</sha_hiDBH>");
      oOut.write("</sha_CutRange>");
      oOut.write("<sha_InitialAge>5000</sha_InitialAge>");
      oOut.write("</SelectionHarvest>");

      oOut.write("</paramFile>");

      oOut.close();
      return sFileName;
  }
}
